"""  TransformationAgent processes transformations found in the transformation database.

The following options can be set for the TransformationAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN TransformationAgent
  :end-before: ##END
  :dedent: 2
  :caption: TransformationAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import os
import datetime
import pickle
import concurrent.futures

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ThreadSafe import Synchronizer
from DIRAC.Core.Utilities.List import breakListIntoChunks, randomize
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.TransformationSystem.Agent.TransformationAgentsUtilities import TransformationAgentsUtilities

__RCSID__ = "$Id$"

AGENT_NAME = "Transformation/TransformationAgent"
gSynchro = Synchronizer()


class TransformationAgent(AgentModule, TransformationAgentsUtilities):
    """Usually subclass of AgentModule"""

    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)
        TransformationAgentsUtilities.__init__(self)

        # few parameters
        self.pluginLocation = ""
        self.transformationStatus = []
        self.maxFiles = 0
        self.transformationTypes = []

        # clients (out of the threads)
        self.transfClient = None

        # parameters for caching
        self.workDirectory = ""
        self.cacheFile = ""
        self.controlDirectory = ""

        self.lastFileOffset = {}
        # Validity of the cache
        self.replicaCache = None
        self.replicaCacheValidity = None
        self.writingCache = False
        self.removedFromCache = 0

        self.noUnusedDelay = 0
        self.unusedFiles = {}
        self.unusedTimeStamp = {}

        self.debug = False
        self.pluginTimeout = {}

    def initialize(self):
        """standard initialize"""
        # few parameters
        self.pluginLocation = self.am_getOption(
            "PluginLocation", "DIRAC.TransformationSystem.Agent.TransformationPlugin"
        )
        self.transformationStatus = self.am_getOption("transformationStatus", ["Active", "Completing", "Flush"])
        # Prepare to change the name of the CS option as MaxFiles is ambiguous
        self.maxFiles = self.am_getOption("MaxFilesToProcess", self.am_getOption("MaxFiles", 5000))

        agentTSTypes = self.am_getOption("TransformationTypes", [])
        if agentTSTypes:
            self.transformationTypes = sorted(agentTSTypes)
        else:
            dataProc = Operations().getValue("Transformations/DataProcessing", ["MCSimulation", "Merge"])
            dataManip = Operations().getValue("Transformations/DataManipulation", ["Replication", "Removal"])
            self.transformationTypes = sorted(dataProc + dataManip)

        # clients
        self.transfClient = TransformationClient()

        # for caching using a pickle file
        self.workDirectory = self.am_getWorkDirectory()
        self.cacheFile = os.path.join(self.workDirectory, "ReplicaCache.pkl")
        self.controlDirectory = self.am_getControlDirectory()

        # remember the offset if any in TS
        self.lastFileOffset = {}

        # Validity of the cache
        self.replicaCache = {}
        self.replicaCacheValidity = self.am_getOption("ReplicaCacheValidity", 2)

        self.noUnusedDelay = self.am_getOption("NoUnusedDelay", 6)

        # Instantiating the ThreadPoolExecutor
        maxNumberOfThreads = self.am_getOption("maxThreadsInPool", 15)
        self.log.info("Multithreaded with %d threads" % maxNumberOfThreads)
        self.threadPoolExecutor = concurrent.futures.ThreadPoolExecutor(max_workers=maxNumberOfThreads)

        self.log.info("Will treat the following transformation types: %s" % str(self.transformationTypes))

        return S_OK()

    def finalize(self):
        """graceful finalization"""

        method = "finalize"
        self._logInfo("Wait for threads to get empty before terminating the agent", method=method)
        self.threadPoolExecutor.shutdown()
        self._logInfo("Threads are empty, terminating the agent...", method=method)
        self.__writeCache()
        return S_OK()

    def execute(self):
        """Just puts transformations in the queue, and spawns threads if there's work to do."""
        # Get the transformations to process
        res = self.getTransformations()
        if not res["OK"]:
            self._logError("Failed to obtain transformations:", res["Message"])
            return S_OK()
        # Process the transformations
        count = 0
        future_to_transID = {}

        for transDict in res["Value"]:
            transID = int(transDict["TransformationID"])
            if transDict.get("InheritedFrom"):
                # Try and move datasets from the ancestor production
                res = self.transfClient.moveFilesToDerivedTransformation(transDict)
                if not res["OK"]:
                    self._logError(
                        "Error moving files from an inherited transformation", res["Message"], transID=transID
                    )
                else:
                    parentProd, movedFiles = res["Value"]
                    if movedFiles:
                        self._logInfo(
                            "Successfully moved files from %d to %d:" % (parentProd, transID), transID=transID
                        )
                        for status, val in movedFiles.items():
                            self._logInfo("\t%d files to status %s" % (val, status), transID=transID)
            count += 1
            future = self.threadPoolExecutor.submit(self._execute, transDict)
            future_to_transID[future] = transID
        self._logInfo("Out of %d transformations, %d put in thread queue" % (len(res["Value"]), count))

        for future in concurrent.futures.as_completed(future_to_transID):
            transID = future_to_transID[future]
            try:
                future.result()
            except Exception as exc:
                self._logError("%d generated an exception: %s" % (transID, exc))
            else:
                self._logInfo("Processed %d" % transID)

        return S_OK()

    def getTransformations(self):
        """Obtain the transformations to be executed - this is executed at the start of every loop (it's really the
        only real thing in the execute()
        """
        transName = self.am_getOption("Transformation", "All")
        method = "getTransformations"
        if transName == "All":
            self._logInfo(
                "Getting all transformations%s, status %s."
                % (
                    " of type %s" % str(self.transformationTypes) if self.transformationTypes else "",
                    str(self.transformationStatus),
                ),
                method=method,
            )
            transfDict = {"Status": self.transformationStatus}
            if self.transformationTypes:
                transfDict["Type"] = self.transformationTypes
            res = self.transfClient.getTransformations(transfDict, extraParams=True)
            if not res["OK"]:
                return res
            transformations = res["Value"]
            self._logInfo("Obtained %d transformations to process" % len(transformations), method=method)
        else:
            self._logInfo("Getting transformation %s." % transName, method=method)
            res = self.transfClient.getTransformation(transName, extraParams=True)
            if not res["OK"]:
                self._logError("Failed to get transformation:", res["Message"], method=method)
                return res
            transformations = [res["Value"]]
        return S_OK(transformations)

    def _getClients(self):
        """returns the clients used in the threads"""
        threadTransformationClient = TransformationClient()
        threadDataManager = DataManager()

        return {"TransformationClient": threadTransformationClient, "DataManager": threadDataManager}

    def _execute(self, transDict):
        """thread - does the real job: processing the transformation to be processed"""

        self._logDebug("Starting _execute")

        # Each thread will have its own clients
        clients = self._getClients()

        try:
            transID = int(transDict["TransformationID"])
            self._logInfo("Processing transformation %s." % transID, transID=transID)
            startTime = time.time()
            res = self.processTransformation(transDict, clients)
            if not res["OK"]:
                self._logInfo("Failed to process transformation:", res["Message"], transID=transID)
        except Exception as x:  # pylint: disable=broad-except
            self._logException("Exception in plugin", lException=x, transID=transID)
        finally:
            if not transID:
                transID = "None"
            self._logInfo("Processed transformation in %.1f seconds" % (time.time() - startTime), transID=transID)

        self._logDebug("Exiting _execute")

    def processTransformation(self, transDict, clients):
        """process a single transformation (in transDict)"""
        method = "processTransformation"
        transID = transDict["TransformationID"]
        forJobs = transDict["Type"].lower() not in ("replication", "removal")

        # First get the LFNs associated to the transformation
        transFiles = self._getTransformationFiles(transDict, clients, replicateOrRemove=not forJobs)
        if not transFiles["OK"]:
            return transFiles
        if not transFiles["Value"]:
            return S_OK()

        if transID not in self.replicaCache:
            self.__readCache(transID)
        transFiles = transFiles["Value"]
        unusedLfns = [f["LFN"] for f in transFiles]
        unusedFiles = len(unusedLfns)

        plugin = transDict.get("Plugin", "Standard")
        # Limit the number of LFNs to be considered for replication or removal as they are treated individually
        if not forJobs:
            maxFiles = Operations().getValue("TransformationPlugins/%s/MaxFilesToProcess" % plugin, 0)
            # Get plugin-specific limit in number of files (0 means no limit)
            totLfns = len(unusedLfns)
            lfnsToProcess = self.__applyReduction(unusedLfns, maxFiles=maxFiles)
            if len(lfnsToProcess) != totLfns:
                self._logInfo(
                    "Reduced number of files from %d to %d" % (totLfns, len(lfnsToProcess)),
                    method=method,
                    transID=transID,
                )
                transFiles = [f for f in transFiles if f["LFN"] in lfnsToProcess]
        else:
            lfnsToProcess = unusedLfns

        # Check the data is available with replicas
        res = self.__getDataReplicas(transDict, lfnsToProcess, clients, forJobs=forJobs)
        if not res["OK"]:
            self._logError("Failed to get data replicas:", res["Message"], method=method, transID=transID)
            return res
        dataReplicas = res["Value"]

        # Get the plug-in type and create the plug-in object
        self._logInfo("Processing transformation with '%s' plug-in." % plugin, method=method, transID=transID)
        res = self.__generatePluginObject(plugin, clients)
        if not res["OK"]:
            return res
        oPlugin = res["Value"]

        # Get the plug-in and set the required params
        oPlugin.setParameters(transDict)
        oPlugin.setInputData(dataReplicas)
        oPlugin.setTransformationFiles(transFiles)
        res = oPlugin.run()
        if not res["OK"]:
            self._logError(
                "Failed to generate tasks for transformation:", res["Message"], method=method, transID=transID
            )
            return res
        tasks = res["Value"]
        self.pluginTimeout[transID] = res.get("Timeout", False)
        # Create the tasks
        allCreated = True
        created = 0
        lfnsInTasks = []
        for se, lfns in tasks:
            res = clients["TransformationClient"].addTaskForTransformation(transID, lfns, se)
            if not res["OK"]:
                self._logError(
                    "Failed to add task generated by plug-in:", res["Message"], method=method, transID=transID
                )
                allCreated = False
            else:
                created += 1
                lfnsInTasks += [lfn for lfn in lfns if lfn in lfnsToProcess]
        if created:
            self._logInfo("Successfully created %d tasks for transformation." % created, method=method, transID=transID)
        else:
            self._logInfo("No new tasks created for transformation.", method=method, transID=transID)
        self.unusedFiles[transID] = unusedFiles - len(lfnsInTasks)
        # If not all files were obtained, move the offset
        lastOffset = self.lastFileOffset.get(transID)
        if lastOffset:
            self.lastFileOffset[transID] = max(0, lastOffset - len(lfnsInTasks))
        self.__removeFilesFromCache(transID, lfnsInTasks)

        # If this production is to Flush
        if transDict["Status"] == "Flush" and allCreated:
            res = clients["TransformationClient"].setTransformationParameter(transID, "Status", "Active")
            if not res["OK"]:
                self._logError(
                    "Failed to update transformation status to 'Active':",
                    res["Message"],
                    method=method,
                    transID=transID,
                )
            else:
                self._logInfo("Updated transformation status to 'Active'.", method=method, transID=transID)
        return S_OK()

    ######################################################################
    #
    # Internal methods used by the agent
    #

    def _getTransformationFiles(self, transDict, clients, statusList=None, replicateOrRemove=False):
        """get the data replicas for a certain transID"""
        # By default, don't skip if no new Unused for DM transformations
        skipIfNoNewUnused = not replicateOrRemove
        transID = transDict["TransformationID"]
        plugin = transDict.get("Plugin", "Standard")
        # Check if files should be sorted and limited in number
        operations = Operations()
        sortedBy = operations.getValue("TransformationPlugins/%s/SortedBy" % plugin, None)
        maxFiles = operations.getValue("TransformationPlugins/%s/MaxFilesToProcess" % plugin, 0)
        # If the NoUnuse delay is explicitly set, we want to take it into account, and skip if no new Unused
        if operations.getValue("TransformationPlugins/%s/NoUnusedDelay" % plugin, 0):
            skipIfNoNewUnused = True
        noUnusedDelay = (
            0
            if self.pluginTimeout.get(transID, False)
            else operations.getValue("TransformationPlugins/%s/NoUnusedDelay" % plugin, self.noUnusedDelay)
        )
        method = "_getTransformationFiles"
        lastOffset = self.lastFileOffset.setdefault(transID, 0)

        # Files that were problematic (either explicit or because SE was banned) may be recovered,
        # and always removing the missing ones
        if not statusList:
            statusList = [TransformationFilesStatus.UNUSED, TransformationFilesStatus.PROB_IN_FC]
        statusList += [TransformationFilesStatus.MISSING_IN_FC] if transDict["Type"] == "Removal" else []
        transClient = clients["TransformationClient"]
        res = transClient.getTransformationFiles(
            condDict={"TransformationID": transID, "Status": statusList},
            orderAttribute=sortedBy,
            offset=lastOffset,
            maxfiles=maxFiles,
        )
        if not res["OK"]:
            self._logError("Failed to obtain input data:", res["Message"], method=method, transID=transID)
            return res
        transFiles = res["Value"]
        if maxFiles and len(transFiles) == maxFiles:
            self.lastFileOffset[transID] += maxFiles
        else:
            del self.lastFileOffset[transID]

        if not transFiles:
            self._logInfo(
                "No '%s' files found for transformation." % ",".join(statusList), method=method, transID=transID
            )
            if transDict["Status"] == "Flush":
                res = transClient.setTransformationParameter(transID, "Status", "Active")
                if not res["OK"]:
                    self._logError(
                        "Failed to update transformation status to 'Active':",
                        res["Message"],
                        method=method,
                        transID=transID,
                    )
                else:
                    self._logInfo("Updated transformation status to 'Active'.", method=method, transID=transID)
            return S_OK()
        # Check if transformation is kicked
        kickFile = os.path.join(self.controlDirectory, "KickTransformation_%s" % str(transID))
        try:
            kickTrans = os.path.exists(kickFile)
            if kickTrans:
                os.remove(kickFile)
        except OSError:
            pass

        # Check if something new happened
        now = datetime.datetime.utcnow()
        if not kickTrans and skipIfNoNewUnused and noUnusedDelay:
            nextStamp = self.unusedTimeStamp.setdefault(transID, now) + datetime.timedelta(hours=noUnusedDelay)
            skip = now < nextStamp
            if len(transFiles) == self.unusedFiles.get(transID, 0) and transDict["Status"] != "Flush" and skip:
                self._logInfo(
                    "No new '%s' files found for transformation." % ",".join(statusList), method=method, transID=transID
                )
                return S_OK()

        self.unusedTimeStamp[transID] = now
        # If files are not Unused, set them Unused
        notUnused = [trFile["LFN"] for trFile in transFiles if trFile["Status"] != TransformationFilesStatus.UNUSED]
        otherStatuses = sorted(
            set([trFile["Status"] for trFile in transFiles]) - set([TransformationFilesStatus.UNUSED])
        )
        if notUnused:
            res = transClient.setFileStatusForTransformation(
                transID, TransformationFilesStatus.UNUSED, notUnused, force=True
            )
            if not res["OK"]:
                self._logError(
                    "Error setting %d files Unused:" % len(notUnused), res["Message"], method=method, transID=transID
                )
            else:
                self._logInfo("Set %d files from %s to Unused" % (len(notUnused), ",".join(otherStatuses)))
                self.__removeFilesFromCache(transID, notUnused)
        return S_OK(transFiles)

    def __applyReduction(self, lfns, maxFiles=None):
        """eventually remove the number of files to be considered"""
        if maxFiles is None:
            maxFiles = self.maxFiles
        if not maxFiles or len(lfns) <= maxFiles:
            return lfns
        return randomize(lfns)[:maxFiles]

    def __getDataReplicas(self, transDict, lfns, clients, forJobs=True):
        """Get the replicas for the LFNs and check their statuses. It first looks within the cache."""
        method = "__getDataReplicas"
        transID = transDict["TransformationID"]
        if "RemoveFile" in transDict["Body"]:
            # When removing files, we don't care about their replicas
            return S_OK(dict.fromkeys(lfns, ["None"]))
        clearCacheFile = os.path.join(self.controlDirectory, "ClearCache_%s" % str(transID))
        try:
            clearCache = os.path.exists(clearCacheFile)
            if clearCache:
                os.remove(clearCacheFile)
        except Exception:
            pass
        if clearCache or transDict["Status"] == "Flush":
            self._logInfo("Replica cache cleared", method=method, transID=transID)
            # We may need to get new replicas
            self.__clearCacheForTrans(transID)
        else:
            # If the cache needs to be cleaned
            self.__cleanCache(transID)
        startTime = time.time()
        dataReplicas = {}
        nLfns = len(lfns)
        self._logVerbose("Getting replicas for %d files" % nLfns, method=method, transID=transID)
        cachedReplicaSets = self.replicaCache.get(transID, {})
        cachedReplicas = {}
        # Merge all sets of replicas
        for replicas in cachedReplicaSets.values():
            cachedReplicas.update(replicas)
        self._logInfo("Number of cached replicas: %d" % len(cachedReplicas), method=method, transID=transID)
        setCached = set(cachedReplicas)
        setLfns = set(lfns)
        for lfn in setLfns & setCached:
            dataReplicas[lfn] = cachedReplicas[lfn]
        newLFNs = setLfns - setCached
        self._logInfo(
            "ReplicaCache hit for %d out of %d LFNs" % (len(dataReplicas), nLfns), method=method, transID=transID
        )
        if newLFNs:
            startTime = time.time()
            self._logInfo("Getting replicas for %d files from catalog" % len(newLFNs), method=method, transID=transID)
            newReplicas = {}
            for chunk in breakListIntoChunks(newLFNs, 10000):
                res = self._getDataReplicasDM(transID, chunk, clients, forJobs=forJobs)
                if res["OK"]:
                    reps = dict((lfn, ses) for lfn, ses in res["Value"].items() if ses)
                    newReplicas.update(reps)
                    self.__updateCache(transID, reps)
                else:
                    self._logWarn(
                        "Failed to get replicas for %d files" % len(chunk),
                        res["Message"],
                        method=method,
                        transID=transID,
                    )

            self._logInfo(
                "Obtained %d replicas from catalog in %.1f seconds" % (len(newReplicas), time.time() - startTime),
                method=method,
                transID=transID,
            )
            dataReplicas.update(newReplicas)
            noReplicas = newLFNs - set(dataReplicas)
            self.__writeCache(transID)
            if noReplicas:
                self._logWarn(
                    "Found %d files without replicas (or only in Failover)" % len(noReplicas),
                    method=method,
                    transID=transID,
                )
        return S_OK(dataReplicas)

    def _getDataReplicasDM(self, transID, lfns, clients, forJobs=True, ignoreMissing=False):
        """Get the replicas for the LFNs and check their statuses, using the replica manager"""
        method = "_getDataReplicasDM"

        startTime = time.time()
        self._logVerbose(
            "Getting replicas%s from catalog for %d files" % (" for jobs" if forJobs else "", len(lfns)),
            method=method,
            transID=transID,
        )
        if forJobs:
            # Get only replicas eligible for jobs
            res = clients["DataManager"].getReplicasForJobs(lfns, getUrl=False)
        else:
            # Get all replicas
            res = clients["DataManager"].getReplicas(lfns, getUrl=False)
        if not res["OK"]:
            return res
        replicas = res["Value"]
        # Prepare a dictionary for all LFNs
        dataReplicas = {}
        self._logVerbose(
            "Replica results for %d files obtained in %.2f seconds" % (len(lfns), time.time() - startTime),
            method=method,
            transID=transID,
        )
        # If files are neither Successful nor Failed, they are set problematic in the FC
        problematicLfns = [lfn for lfn in lfns if lfn not in replicas["Successful"] and lfn not in replicas["Failed"]]
        if problematicLfns:
            self._logInfo("%d files found problematic in the catalog, set ProbInFC" % len(problematicLfns))
            res = clients["TransformationClient"].setFileStatusForTransformation(
                transID, TransformationFilesStatus.PROB_IN_FC, problematicLfns
            )
            if not res["OK"]:
                self._logError(
                    "Failed to update status of problematic files:", res["Message"], method=method, transID=transID
                )
        # Create a dictionary containing all the file replicas
        failoverLfns = []
        for lfn, replicaDict in replicas["Successful"].items():
            for se in replicaDict:
                # This remains here for backward compatibility in case VOs have not defined SEs not to be used for jobs
                if forJobs and "failover" in se.lower():
                    self._logVerbose("Ignoring failover replica for %s." % lfn, method=method, transID=transID)
                else:
                    dataReplicas.setdefault(lfn, []).append(se)
            if not dataReplicas.get(lfn):
                failoverLfns.append(lfn)
        if failoverLfns:
            self._logVerbose("%d files have no replica but possibly in Failover SE" % len(failoverLfns))
        # Make sure that file missing from the catalog are marked in the transformation DB.
        missingLfns = []
        for lfn, reason in replicas["Failed"].items():
            if "No such file or directory" in reason:
                self._logVerbose("%s not found in the catalog." % lfn, method=method, transID=transID)
                missingLfns.append(lfn)
        if missingLfns:
            self._logInfo("%d files not found in the catalog" % len(missingLfns))
            if ignoreMissing:
                dataReplicas.update(dict.fromkeys(missingLfns, []))
            else:
                res = clients["TransformationClient"].setFileStatusForTransformation(
                    transID, TransformationFilesStatus.MISSING_IN_FC, missingLfns
                )
                if not res["OK"]:
                    self._logError(
                        "Failed to update status of missing files:", res["Message"], method=method, transID=transID
                    )
        return S_OK(dataReplicas)

    def __updateCache(self, transID, newReplicas):
        """Add replicas to the cache"""
        self.replicaCache.setdefault(transID, {})[datetime.datetime.utcnow()] = newReplicas

    def __clearCacheForTrans(self, transID):
        """Remove all replicas for a transformation"""
        self.replicaCache.pop(transID, None)

    def __cleanReplicas(self, transID, lfns):
        """Remove cached replicas that are not in a list"""
        cachedReplicas = set()
        for replicas in self.replicaCache.get(transID, {}).values():
            cachedReplicas.update(replicas)
        toRemove = cachedReplicas - set(lfns)
        if toRemove:
            self._logInfo("Remove %d files from cache" % len(toRemove), method="__cleanReplicas", transID=transID)
            self.__removeFromCache(transID, toRemove)

    def __cleanCache(self, transID):
        """Cleans the cache"""
        try:
            if transID in self.replicaCache:
                timeLimit = datetime.datetime.utcnow() - datetime.timedelta(days=self.replicaCacheValidity)
                for updateTime in set(self.replicaCache[transID]):
                    nCache = len(self.replicaCache[transID][updateTime])
                    if updateTime < timeLimit or not nCache:
                        self._logInfo(
                            "Clear %s replicas for transformation %s, time %s"
                            % ("%d cached" % nCache if nCache else "empty cache", str(transID), str(updateTime)),
                            transID=transID,
                            method="__cleanCache",
                        )
                        del self.replicaCache[transID][updateTime]
                # Remove empty transformations
                if not self.replicaCache[transID]:
                    del self.replicaCache[transID]
        except Exception as x:
            self._logException("Exception when cleaning replica cache:", lException=x)

    def __removeFilesFromCache(self, transID, lfns):
        removed = self.__removeFromCache(transID, lfns)
        if removed:
            self._logInfo("Removed %d replicas from cache" % removed, method="__removeFilesFromCache", transID=transID)
            self.__writeCache(transID)

    def __removeFromCache(self, transID, lfns):
        if transID not in self.replicaCache:
            return
        removed = 0
        if self.replicaCache[transID] and lfns:
            for lfn in lfns:
                for timeKey in self.replicaCache[transID]:
                    if self.replicaCache[transID][timeKey].pop(lfn, None):
                        removed += 1
        return removed

    def __cacheFile(self, transID):
        return self.cacheFile.replace(".pkl", "_%s.pkl" % str(transID))

    @gSynchro
    def __readCache(self, transID):
        """Reads from the cache"""
        if transID in self.replicaCache:
            return
        try:
            method = "__readCache"
            fileName = self.__cacheFile(transID)
            if not os.path.exists(fileName):
                self.replicaCache[transID] = {}
            else:
                with open(fileName, "rb") as cacheFile:
                    self.replicaCache[transID] = pickle.load(cacheFile)
                self._logInfo(
                    "Successfully loaded replica cache from file %s (%d files)"
                    % (fileName, self.__filesInCache(transID)),
                    method=method,
                    transID=transID,
                )
        except Exception as x:
            self._logException(
                "Failed to load replica cache from file %s" % fileName, lException=x, method=method, transID=transID
            )
            self.replicaCache[transID] = {}

    def __filesInCache(self, transID):
        cache = self.replicaCache.get(transID, {})
        return sum(len(lfns) for lfns in cache.values())

    @gSynchro
    def __writeCache(self, transID=None):
        """Writes the cache"""
        method = "__writeCache"
        try:
            startTime = time.time()
            transList = [transID] if transID else set(self.replicaCache)
            filesInCache = 0
            nCache = 0
            for t_id in transList:
                # Protect the copy of the cache
                filesInCache += self.__filesInCache(t_id)
                # write to a temporary file in order to avoid corrupted files
                cacheFile = self.__cacheFile(t_id)
                tmpFile = cacheFile + ".tmp"
                with open(tmpFile, "wb") as fd:
                    pickle.dump(self.replicaCache.get(t_id, {}), fd)
                # Now rename the file as it shold
                os.rename(tmpFile, cacheFile)
                nCache += 1
            self._logInfo(
                "Successfully wrote %d replica cache file(s) (%d files) in %.1f seconds"
                % (nCache, filesInCache, time.time() - startTime),
                method=method,
                transID=transID if transID else None,
            )
        except Exception as x:
            self._logException(
                "Could not write replica cache file %s" % cacheFile, lException=x, method=method, transID=t_id
            )

    def __generatePluginObject(self, plugin, clients):
        """This simply instantiates the TransformationPlugin class with the relevant plugin name"""
        try:
            plugModule = __import__(self.pluginLocation, globals(), locals(), ["TransformationPlugin"])
        except ImportError as e:
            self._logException(
                "Failed to import 'TransformationPlugin' %s" % plugin, lException=e, method="__generatePluginObject"
            )
            return S_ERROR()
        try:
            plugin_o = getattr(plugModule, "TransformationPlugin")(
                "%s" % plugin, transClient=clients["TransformationClient"], dataManager=clients["DataManager"]
            )
            return S_OK(plugin_o)
        except AttributeError as e:
            self._logException("Failed to create %s()" % plugin, lException=e, method="__generatePluginObject")
            return S_ERROR()
        plugin_o.setDirectory(self.workDirectory)
        plugin_o.setCallback(self.pluginCallback)

    def pluginCallback(self, transID, invalidateCache=False):
        """Standard plugin callback"""
        if invalidateCache:
            try:
                if transID in self.replicaCache:
                    self._logInfo(
                        "Removed cached replicas for transformation", method="pluginCallBack", transID=transID
                    )
                    self.replicaCache.pop(transID)
                    self.__writeCache(transID)
            except Exception:
                pass



""" GGUSTicketsClient class is a client for the GGUS Tickets DB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six.moves.urllib_error import URLError

from suds import WebFault
from suds.client import Client

from DIRAC import gConfig, S_OK, S_ERROR

__RCSID__ = "$Id$"


def getGGUSURL(vo=None, siteName=None):
    """create the URL to get tickets relative to the site ( opened only ! ):"""
    if vo is None:
        vo = gConfig.getOption("DIRAC/VirtualOrganization")

    ggusURL = "https://ggus.eu/ws/ticket_search.php?show_columns_check[]=REQUEST_ID&"
    ggusURL += "show_columns_check[]=TICKET_TYPE&show_columns_check[]=AFFECTED_VO&show_columns_check[]="
    ggusURL += "AFFECTED_SITE&show_columns_check[]=PRIORITY&show_columns_check[]=RESPONSIBLE_UNIT&show_"
    ggusURL += "columns_check[]=STATUS&show_columns_check[]=DATE_OF_CREATION&show_columns_check[]=LAST_UPDATE&"
    ggusURL += "show_columns_check[]=TYPE_OF_PROBLEM&show_columns_check[]=SUBJECT&ticket=&supportunit=all&su_"
    ggusURL += "hierarchy=all&vo=%s&user=&keyword=&involvedsupporter=&assignto=" % vo
    if siteName:
        ggusURL += "&affectedsite=%s" % siteName
    ggusURL += "&specattrib=0&status=open&priority=all&typeofproblem=all&ticketcategory=&mouarea=&technology_"
    ggusURL += "provider=&date_type=creation+date&radiotf=1&timeframe=any&untouched_date=&orderticketsby=GHD_"
    ggusURL += "INT_REQUEST_ID&orderhow=descending"

    return ggusURL


class GGUSTicketsClient(object):
    def __init__(self):

        # create client instance using GGUS wsdl:
        self.gclient = Client("https://prod-ars.ggus.eu/arsys/WSDL/public/prod-ars/GGUS")
        authInfo = self.gclient.factory.create("AuthenticationInfo")
        authInfo.userName = "ticketinfo"
        authInfo.password = "TicketInfo"
        self.vo = gConfig.getOption("DIRAC/VirtualOrganization")
        self.gclient.set_options(soapheaders=authInfo)

    ################################################################################

    def getTicketsList(self, siteName=None, startDate=None, endDate=None):
        """Return tickets of entity in name

        :param name: should be the name of the site
        :param startDate: starting date (optional)
        :param endDate: end date (optional)
        """

        # prepare the query string:

        # query = '\'GHD_Affected Site\'=\"' + siteName + '\" AND \'GHD_Affected VO\'="lhcb"'

        query = "'GHD_Affected VO'=%s" % self.vo
        if siteName is not None:
            query += " AND 'GHD_Affected Site'=\"" + siteName + '"'

        if startDate is not None:
            query = query + " AND 'GHD_Date Of Creation'>" + str(startDate)

        if endDate is not None:
            query = query + " AND 'GHD_Date Of Creation'<" + str(endDate)

        # the query must be into a try block. Empty queries, though formally correct, raise an exception
        try:
            ticketList = self.gclient.service.TicketGetList(query)
        except WebFault as e:
            return S_ERROR(e)
        except URLError as e:
            return S_ERROR(e)

        return self.globalStatistics(ticketList)

    ################################################################################

    def globalStatistics(self, ticketList):
        """
        Get some statistics about the tickets for the site: total number
        of tickets and number of ticket in different status
        """

        # initialize the dictionary of tickets to return
        selectedTickets = {}

        # openStates = [ 'assigned', 'in progress', 'new', 'on hold', 'reopened', 'waiting for reply' ]
        terminalStates = ["solved", "unsolved", "verified", "closed"]

        for ticket in ticketList:

            _id = str(ticket.GHD_Request_ID)
            _status = str(ticket.GHD_Status)
            _shortDescription = str(ticket.GHD_Short_Description)
            _priority = str(ticket.GHD_Priority)
            if not hasattr(ticket, "GHD_Affected_Site"):
                continue
            site = str(ticket.GHD_Affected_Site)

            # We do not want closed tickets
            if _status in terminalStates:
                continue

            if site not in selectedTickets:
                selectedTickets[site] = {"URL": getGGUSURL(self.vo, site)}

            if _priority not in selectedTickets[site]:
                selectedTickets[site][_priority] = []

            selectedTickets[site][_priority].append((_id, _shortDescription, _status))

        return S_OK(selectedTickets)



""" Here, we need some documentation...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import threading
import time
import select
import socket

from concurrent.futures import ThreadPoolExecutor

try:
    import selectors
except ImportError:
    import selectors2 as selectors

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.private.TransportPool import getGlobalTransportPool
from DIRAC.Core.Utilities.ReturnValues import isReturnStructure
from DIRAC.Core.DISET.private.MessageFactory import MessageFactory, DummyMessage


class MessageBroker(object):
    def __init__(self, name, transportPool=None, threadPool=None):
        self.__name = name
        self.__messageTransports = {}
        self.__msgCounter = 0
        self.__msgCounterLock = threading.Lock()
        self.__responseCallbacks = {}
        self.__msgInTransport = {}
        self.__listenPersistConn = False
        self.__useMessageObjects = True
        self.__callbacksLock = threading.Condition()
        self.__trInOutLock = threading.Lock()
        self.__msgFactory = MessageFactory()
        self.__log = gLogger.getSubLogger("MSGBRK")
        if not transportPool:
            transportPool = getGlobalTransportPool()
        self.__trPool = transportPool
        if not threadPool:
            threadPool = ThreadPoolExecutor(100)
        self.__threadPool = threadPool
        self.__listeningForMessages = False
        self.__listenThread = None

    def getNumConnections(self):
        return len(self.__messageTransports)

    def getMsgFactory(self):
        return self.__msgFactory

    def useMessageObjects(self, bD):
        self.__useMessageObjects = bD

    # Message id generation

    def __generateMsgId(self):
        self.__msgCounterLock.acquire()
        try:
            msgId = "%s:%d" % (self.__name, self.__msgCounter)
            self.__msgCounter += 1
            return msgId
        finally:
            self.__msgCounterLock.release()

    def getTransportPool(self):
        return self.__trPool

    # Add and remove transport to/from broker

    def addTransport(self, transport, *args, **kwargs):
        trid = self.__trPool.add(transport)
        try:
            result = self.addTransportId(trid, *args, **kwargs)
        except Exception as e:
            gLogger.exception("Cannot add transport id", lException=e)
            result = S_ERROR("Cannot add transport id")
        if not result["OK"]:
            self.__trPool.remove(trid)
            return result
        return S_OK(trid)

    def addTransportId(
        self,
        trid,
        svcName,
        receiveMessageCallback=None,
        disconnectCallback=None,
        idleRead=False,
        listenToConnection=True,
    ):
        self.__trInOutLock.acquire()
        try:
            if trid in self.__messageTransports:
                return S_OK()
            tr = self.__trPool.get(trid)
            if not tr:
                return S_ERROR("No transport with id %s registered" % trid)
            self.__messageTransports[trid] = {
                "transport": tr,
                "svcName": svcName,
                "cbReceiveMessage": receiveMessageCallback,
                "cbDisconnect": disconnectCallback,
                "listen": listenToConnection,
                "idleRead": idleRead,
            }
            self.__startListeningThread()
            return S_OK()
        finally:
            self.__trInOutLock.release()

    def listenToTransport(self, trid, listen=True):
        self.__trInOutLock.acquire()
        try:
            if trid in self.__messageTransports:
                self.__messageTransports[trid]["listen"] = listen
            self.__startListeningThread()
        finally:
            self.__trInOutLock.release()

    # Listen to connections

    def __startListeningThread(self):
        threadDead = (
            self.__listeningForMessages and self.__listenThread is not None and not self.__listenThread.is_alive()
        )
        if not self.__listeningForMessages or threadDead:
            self.__listeningForMessages = True
            self.__listenThread = threading.Thread(target=self.__listenAutoReceiveConnections)
            self.__listenThread.setDaemon(True)
            self.__listenThread.start()

    def __listenAutoReceiveConnections(self):
        while self.__listeningForMessages:
            self.__trInOutLock.acquire()
            try:
                # TODO: A single DefaultSelector instance can probably be shared by all threads
                sel = selectors.DefaultSelector()
                for trid in self.__messageTransports:
                    mt = self.__messageTransports[trid]
                    if not mt["listen"]:
                        continue
                    sel.register(mt["transport"].getSocket(), selectors.EVENT_READ, trid)
                if not sel.get_map():
                    self.__listeningForMessages = False
                    return
            finally:
                self.__trInOutLock.release()

            try:
                events = sel.select(timeout=1)
            except Exception:
                gLogger.exception("Exception while selecting persistent connections")
                continue

            for key, event in events:
                if event & selectors.EVENT_READ:
                    trid = key.data
                    if trid in self.__messageTransports:
                        result = self.__receiveMsgDataAndQueue(trid)
                        if not result["OK"]:
                            self.removeTransport(trid)

    # Process received data functions

    def __receiveMsgDataAndQueue(self, trid):
        # Receive
        result = self.__trPool.receive(
            trid, blockAfterKeepAlive=False, idleReceive=self.__messageTransports[trid]["idleRead"]
        )
        self.__log.debug("[trid %s] Received data: %s" % (trid, str(result)))
        # If error close transport and exit
        if not result["OK"]:
            self.__log.debug("[trid %s] ERROR RCV DATA %s" % (trid, result["Message"]))
            gLogger.warn(
                "Error while receiving message",
                "from %s : %s" % (self.__trPool.get(trid).getFormattedCredentials(), result["Message"]),
            )
            return self.removeTransport(trid)
        self.__threadPool.submit(self.__processIncomingData, trid, result)
        return S_OK()

    def __processIncomingData(self, trid, receivedResult):
        # If keep alive, return OK
        if receivedResult.get("keepAlive"):
            return S_OK()
        # If idle read return
        self.__trInOutLock.acquire()
        try:
            idleRead = self.__messageTransports[trid]["idleRead"]
        except KeyError:
            return S_ERROR("Transport %s unknown" % trid)
        finally:
            self.__trInOutLock.release()
        if idleRead:
            if receivedResult["Value"]:
                gLogger.fatal("OOOops. Idle read has returned data!")
            return S_OK()
        if not receivedResult["Value"]:
            self.__log.debug("Transport %s closed connection" % trid)
            return self.removeTransport(trid)
        # This is a message req/resp
        msg = receivedResult["Value"]
        # Valid message?
        if "request" not in msg:
            gLogger.warn("Received data does not seem to be a message !!!!")
            return self.removeTransport(trid)
        # Decide if it's a response or a request
        if msg["request"]:
            # If message has Id return ACK to received
            if "id" in msg:
                self.__sendResponse(trid, msg["id"], S_OK())
            # Process msg
            result = self.__processIncomingRequest(trid, msg)
        else:
            result = self.__processIncomingResponse(trid, msg)
        # If error close the transport
        if not result["OK"]:
            gLogger.info("Closing transport because of error while processing message", result["Message"])
            return self.removeTransport(trid)
        return S_OK()

    def __processIncomingRequest(self, trid, msg):
        self.__trInOutLock.acquire()
        try:
            rcvCB = self.__messageTransports[trid]["cbReceiveMessage"]
        except KeyError:
            return S_ERROR("Transport %s unknown" % trid)
        finally:
            self.__trInOutLock.release()
        if not rcvCB:
            gLogger.fatal("Transport %s does not have a callback defined and a message arrived!" % trid)
            return S_ERROR("No message was expected in for this transport")
        # Check message has id and name
        for requiredField in ["name"]:
            if requiredField not in msg:
                gLogger.error("Message does not have required field", requiredField)
                return S_ERROR("Message does not have %s" % requiredField)
        # Load message
        if "attrs" in msg:
            attrs = msg["attrs"]
            if not isinstance(attrs, (tuple, list)):
                return S_ERROR("Message args has to be a tuple or a list, not %s" % type(attrs))
        else:
            attrs = None
        # Do we "unpack" or do we send the raw data to the callback?
        if self.__useMessageObjects:
            result = self.__msgFactory.createMessage(self.__messageTransports[trid]["svcName"], msg["name"], attrs)
            if not result["OK"]:
                return result
            msgObj = result["Value"]
        else:
            msgObj = DummyMessage(msg)
        # Is msg ok?
        if not msgObj.isOK():
            return S_ERROR("Messsage is invalid")
        try:
            # Callback it and return response
            result = rcvCB(trid, msgObj)
            if not isReturnStructure(result):
                return S_ERROR("Request function does not return a result structure")
            return result
        except Exception as e:
            # Whoops. Show exception and return
            gLogger.exception("Exception while processing message %s" % msg["name"], lException=e)
            return S_ERROR("Exception while processing message %s: %s" % (msg["name"], str(e)))

    def __processIncomingResponse(self, trid, msg):
        # This is a message response
        for requiredField in ("id", "result"):
            if requiredField not in msg:
                gLogger.error("Message does not have required field", requiredField)
                return S_ERROR("Message does not have %s" % requiredField)
        if not isReturnStructure(msg["result"]):
            return S_ERROR("Message response did not return a result structure")
        return self.__notifyCallback(msg["id"], msg["result"])

    # Sending functions

    def __sendResponse(self, trid, msgId, msgResult):
        msgResponse = {"request": False, "id": msgId, "result": msgResult}
        self.__trPool.send(trid, S_OK(msgResponse))

    def sendMessage(self, trid, msgObj):
        if not msgObj.isOK():
            return S_ERROR("Message is not ready to be sent")
        result = self.__sendMessage(trid, msgObj)
        if not result["OK"]:
            self.removeTransport(trid)
        return result

    def __sendMessage(self, trid, msgObj):
        if not self.__trPool.exists(trid):
            return S_ERROR("Not transport with id %s defined for messaging" % trid)

        msg = {"request": True, "name": msgObj.getName()}
        attrs = msgObj.dumpAttrs()["Value"]
        msg["attrs"] = attrs
        waitForAck = msgObj.getWaitForAck()

        if not waitForAck:
            return self.__trPool.send(trid, S_OK(msg))

        msgId = self.__generateMsgId()
        msg["id"] = msgId

        self.__generateMessageResponse(trid, msgId)
        result = self.__trPool.send(trid, S_OK(msg))

        # Lock and generate and wait
        self.__callbacksLock.acquire()
        try:
            if not result["OK"]:
                # Release lock and exit
                self.__clearCallback(msgId)
                return result

            return self.__waitForMessageResponse(msgId)
        finally:
            self.__callbacksLock.release()

    # Callback nightmare

    # Lock need to have been aquired prior to func
    def __generateMessageResponse(self, trid, msgId):
        self.__callbacksLock.acquire()
        try:
            if msgId in self.__responseCallbacks:
                return self.__responseCallbacks[msgId]
            if trid not in self.__msgInTransport:
                self.__msgInTransport[trid] = set()
            self.__msgInTransport[trid].add(msgId)
            self.__responseCallbacks[msgId] = {"creationTime": time.time(), "trid": trid}
            return self.__responseCallbacks[msgId]
        finally:
            self.__callbacksLock.release()

    # Lock need to have been aquired prior to func
    def __waitForMessageResponse(self, msgId):
        if msgId not in self.__responseCallbacks:
            return S_ERROR("Invalid msg id")
        respCallback = self.__responseCallbacks[msgId]
        while "result" not in respCallback and time.time() - respCallback["creationTime"] < 30:
            self.__callbacksLock.wait(30)
        self.__clearCallback(msgId)
        if "result" in respCallback:
            return respCallback["result"]
        return S_ERROR("Timeout while waiting for message ack")

    def __clearCallback(self, msgId):
        if msgId not in self.__responseCallbacks:
            return False
        trid = self.__responseCallbacks[msgId]["trid"]
        self.__responseCallbacks.pop(msgId)
        try:
            self.__msgInTransport[trid].remove(msgId)
        except KeyError:
            pass
        return True

    # Lock need to have been aquired prior to func
    def __setCallbackResult(self, msgId, result=False):
        if msgId not in self.__responseCallbacks:
            return False
        self.__responseCallbacks[msgId]["result"] = result
        return True

    def __notifyCallback(self, msgId, msgResult):
        self.__callbacksLock.acquire()
        try:
            if self.__setCallbackResult(msgId, msgResult):
                self.__callbacksLock.notifyAll()
        finally:
            self.__callbacksLock.release()
        return S_OK()

    def removeTransport(self, trid, closeTransport=True):
        # Delete from the message Transports
        self.__trInOutLock.acquire()
        try:
            if trid not in self.__messageTransports:
                return S_OK()

            # Save the disconnect callback if it's there
            if self.__messageTransports[trid]["cbDisconnect"]:
                cbDisconnect = self.__messageTransports[trid]["cbDisconnect"]
            else:
                cbDisconnect = False

            self.__messageTransports.pop(trid)
            if closeTransport:
                self.__trPool.close(trid)
        finally:
            self.__trInOutLock.release()

        # Flush remaining messages
        self.__callbacksLock.acquire()
        try:
            msgIds = False
            if trid in self.__msgInTransport:
                msgIds = set(self.__msgInTransport[trid])
                self.__msgInTransport.pop(trid)
                for msgId in msgIds:
                    self.__setCallbackResult(msgId, S_ERROR("Connection closed by peer"))
            self.__callbacksLock.notifyAll()
        finally:
            self.__callbacksLock.release()

        # Queue the disconnect CB if it's there
        if cbDisconnect:
            self.__threadPool.submit(cbDisconnect, trid)

        return S_OK()


class MessageSender(object):
    def __init__(self, serviceName, msgBroker):
        self.__serviceName = serviceName
        self.__msgBroker = msgBroker

    def getServiceName(self):
        return self.__serviceName

    def sendMessage(self, trid, msgObj):
        return self.__msgBroker.sendMessage(trid, msgObj)

    def createMessage(self, msgName):
        return self.__msgBroker.__msgFactory.createMessage(self.__serviceName, msgName)


gMessageBroker = False


def getGlobalMessageBroker():
    global gMessageBroker
    if not gMessageBroker:
        gMessageBroker = MessageBroker("GlobalMessageBroker", transportPool=getGlobalTransportPool())
    return gMessageBroker



""" ARC Computing Element
    Using the ARC API now

**Configuration Parameters**

Configuration for the ARCComputingElement submission can be done via the configuration system. See the page about
configuring :ref:`resourcesComputing` for where the options can be placed.

XRSLExtraString:
   Default additional string for ARC submit files. Should be written in the following format::

     (key = "value")

XRSLMPExtraString:
   Default additional string for ARC submit files for multi-processor jobs. Should be written in the following format::

     (key = "value")

Host:
   The host for the ARC CE, used to overwrite the CE name.

WorkingDirectory:
   Directory where the pilot log files are stored locally. For instance::

     /opt/dirac/pro/runit/WorkloadManagement/SiteDirectorArc

EndpointType:
   Name of the protocol to use to interact with ARC services: Emies and Gridftp are supported.
   Gridftp communicates with gridftpd services providing authentication and encryption for file transfers.
   ARC developers are going to drop it in the future.
   Emies is another protocol that allows to interact with A-REX services that provide additional features
   (support of OIDC tokens).

**Code Documentation**
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import os
import stat

import arc  # Has to work if this module is called #pylint: disable=import-error
from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping
from DIRAC.Core.Utilities.Subprocess import shellCall
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
from DIRAC.Resources.Computing.ComputingElement import ComputingElement
from DIRAC.WorkloadManagementSystem.Client import PilotStatus

# Uncomment the following 5 lines for getting verbose ARC api output (debugging)
# import sys
# logstdout = arc.LogStream(sys.stdout)
# logstdout.setFormat(arc.ShortFormat)
# arc.Logger_getRootLogger().addDestination(logstdout)
# arc.Logger_getRootLogger().setThreshold(arc.VERBOSE)

CE_NAME = "ARC"
MANDATORY_PARAMETERS = ["Queue"]  # Mandatory for ARC CEs in GLUE2?
STATES_MAP = {
    "Accepted": PilotStatus.WAITING,
    "Preparing": PilotStatus.WAITING,
    "Submitting": PilotStatus.WAITING,
    "Queuing": PilotStatus.WAITING,
    "Undefined": PilotStatus.UNKNOWN,
    "Running": PilotStatus.RUNNING,
    "Finishing": PilotStatus.RUNNING,
    "Deleted": PilotStatus.ABORTED,
    "Killed": PilotStatus.ABORTED,
    "Failed": PilotStatus.FAILED,
    "Hold": PilotStatus.FAILED,
    "Finished": PilotStatus.DONE,
    "Other": PilotStatus.DONE,
}


class ARCComputingElement(ComputingElement):

    #############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(ARCComputingElement, self).__init__(ceUniqueID)

        self.ceType = CE_NAME
        self.submittedJobs = 0
        self.mandatoryParameters = MANDATORY_PARAMETERS
        self.pilotProxy = ""
        self.queue = ""
        self.gridEnv = ""
        self.ceHost = self.ceName
        self.endpointType = "Gridftp"
        self.usercfg = arc.common.UserConfig()
        # set the timeout to the default 20 seconds in case the UserConfig constructor did not
        self.usercfg.Timeout(20)  # pylint: disable=pointless-statement
        self.ceHost = self.ceParameters.get("Host", self.ceName)
        self.gridEnv = self.ceParameters.get("GridEnv", self.gridEnv)

        # ARC endpoint types (Gridftp, Emies)
        endpointType = self.ceParameters.get("EndpointType", self.endpointType)
        if endpointType not in ["Gridftp", "Emies"]:
            self.log.warn("Unknown ARC endpoint, change to default", self.endpointType)
        else:
            self.endpointType = endpointType

        # Used in getJobStatus
        self.mapStates = STATES_MAP
        # Do these after all other initialisations, in case something barks
        self.xrslExtraString = self.__getXRSLExtraString()
        self.xrslMPExtraString = self.__getXRSLExtraString(multiprocessor=True)

    #############################################################################

    def __getARCJob(self, jobID):
        """Create an ARC Job with all the needed / possible parameters defined.
        By the time we come here, the environment variable X509_USER_PROXY should already be set
        """
        j = arc.Job()
        j.JobID = str(jobID)
        j.IDFromEndpoint = os.path.basename(j.JobID)

        if self.endpointType == "Gridftp":
            statURL = "ldap://%s:2135/Mds-Vo-Name=local,o=grid??sub?(nordugrid-job-globalid=%s)" % (self.ceHost, jobID)
            j.JobStatusURL = arc.URL(str(statURL))
            j.JobStatusInterfaceName = "org.nordugrid.ldapng"

            mangURL = "gsiftp://%s:2811/jobs/" % (self.ceHost)
            j.JobManagementURL = arc.URL(str(mangURL))
            j.JobManagementInterfaceName = "org.nordugrid.gridftpjob"

            j.ServiceInformationURL = j.JobManagementURL
            j.ServiceInformationInterfaceName = "org.nordugrid.ldapng"
        else:
            commonURL = "https://%s:8443/arex" % self.ceHost
            j.JobStatusURL = arc.URL(str(commonURL))
            j.JobStatusInterfaceName = "org.ogf.glue.emies.activitymanagement"

            j.JobManagementURL = arc.URL(str(commonURL))
            j.JobManagementInterfaceName = "org.ogf.glue.emies.activitymanagement"

            j.ServiceInformationURL = arc.URL(str(commonURL))
            j.ServiceInformationInterfaceName = "org.ogf.glue.emies.resourceinfo"

        j.PrepareHandler(self.usercfg)
        return j

    def __getXRSLExtraString(self, multiprocessor=False):
        # For the XRSL additional string from configuration - only done at initialisation time
        # If this string changes, the corresponding (ARC) site directors have to be restarted
        #
        # Variable = XRSLExtraString (or XRSLMPExtraString for multi processor mode)
        # Default value = ''
        #   If you give a value, I think it should be of the form
        #          (aaa = "xxx")
        #   Otherwise the ARC job description parser will have a fit
        # Locations searched in order :
        # Top priority    : Resources/Sites/<Grid>/<Site>/CEs/<CE>/XRSLExtraString
        # Second priority : Resources/Sites/<Grid>/<Site>/XRSLExtraString
        # Default         : Resources/Computing/CEDefaults/XRSLExtraString
        #
        xrslExtraString = ""  # Start with the default value
        result = getCESiteMapping(self.ceHost)
        if not result["OK"] or not result["Value"]:
            self.log.error("Unknown CE ...")
            return
        self.site = result["Value"][self.ceHost]
        # Now we know the site. Get the grid
        grid = self.site.split(".")[0]
        # The different possibilities that we have agreed upon
        if multiprocessor:
            xtraVariable = "XRSLMPExtraString"
        else:
            xtraVariable = "XRSLExtraString"
        firstOption = "Resources/Sites/%s/%s/CEs/%s/%s" % (grid, self.site, self.ceHost, xtraVariable)
        secondOption = "Resources/Sites/%s/%s/%s" % (grid, self.site, xtraVariable)
        defaultOption = "Resources/Computing/CEDefaults/%s" % xtraVariable
        # Now go about getting the string in the agreed order
        self.log.debug("Trying to get %s : first option %s" % (xtraVariable, firstOption))
        result = gConfig.getValue(firstOption, defaultValue="")
        if result != "":
            xrslExtraString = result
            self.log.debug("Found %s : %s" % (xtraVariable, xrslExtraString))
        else:
            self.log.debug("Trying to get %s : second option %s" % (xtraVariable, secondOption))
            result = gConfig.getValue(secondOption, defaultValue="")
            if result != "":
                xrslExtraString = result
                self.log.debug("Found %s : %s" % (xtraVariable, xrslExtraString))
            else:
                self.log.debug("Trying to get %s : default option %s" % (xtraVariable, defaultOption))
                result = gConfig.getValue(defaultOption, defaultValue="")
                if result != "":
                    xrslExtraString = result
                    self.log.debug("Found %s : %s" % (xtraVariable, xrslExtraString))
        if xrslExtraString:
            self.log.always("%s : %s" % (xtraVariable, xrslExtraString))
            self.log.always(" --- to be added to pilots going to CE : %s" % self.ceHost)
        return xrslExtraString

    #############################################################################
    def _addCEConfigDefaults(self):
        """Method to make sure all necessary Configuration Parameters are defined"""
        # First assure that any global parameters are loaded
        ComputingElement._addCEConfigDefaults(self)

    #############################################################################
    def __writeXRSL(self, executableFile):
        """Create the JDL for submission"""
        diracStamp = makeGuid()[:8]
        # Evaluate the number of processors to allocate
        nProcessors = self.ceParameters.get("NumberOfProcessors", 1)

        xrslMPAdditions = ""
        if nProcessors and nProcessors > 1:
            xrslMPAdditions = """
(count = %(processors)u)
(countpernode = %(processorsPerNode)u)
%(xrslMPExtraString)s
      """ % {
                "processors": nProcessors,
                "processorsPerNode": nProcessors,  # This basically says that we want all processors on the same node
                "xrslMPExtraString": self.xrslMPExtraString,
            }

        xrsl = """
&(executable="%(executable)s")
(inputFiles=(%(executable)s "%(executableFile)s"))
(stdout="%(diracStamp)s.out")
(stderr="%(diracStamp)s.err")
(outputFiles=("%(diracStamp)s.out" "") ("%(diracStamp)s.err" ""))
(queue=%(queue)s)
%(xrslMPAdditions)s
%(xrslExtraString)s
    """ % {
            "executableFile": executableFile,
            "executable": os.path.basename(executableFile),
            "diracStamp": diracStamp,
            "queue": self.arcQueue,
            "xrslMPAdditions": xrslMPAdditions,
            "xrslExtraString": self.xrslExtraString,
        }

        return xrsl, diracStamp

    #############################################################################
    def _reset(self):
        self.queue = self.ceParameters.get("CEQueueName", self.ceParameters["Queue"])
        if "GridEnv" in self.ceParameters:
            self.gridEnv = self.ceParameters["GridEnv"]
        return S_OK()

    #############################################################################
    def submitJob(self, executableFile, proxy, numberOfJobs=1):
        """Method to submit job"""

        # Assume that the ARC queues are always of the format nordugrid-<batchSystem>-<queue>
        # And none of our supported batch systems have a "-" in their name
        self.arcQueue = self.queue.split("-", 2)[2]
        result = self._prepareProxy()
        if not result["OK"]:
            self.log.error("ARCComputingElement: failed to set up proxy", result["Message"])
            return result
        self.usercfg.ProxyPath(os.environ["X509_USER_PROXY"])

        self.log.verbose("Executable file path: %s" % executableFile)
        if not os.access(executableFile, 5):
            os.chmod(executableFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH + stat.S_IXOTH)

        batchIDList = []
        stampDict = {}

        if self.endpointType == "Gridftp":
            endpoint = arc.Endpoint(str(self.ceHost + ":2811/jobs"), arc.Endpoint.JOBSUBMIT, "org.nordugrid.gridftpjob")
        else:
            endpoint = arc.Endpoint(
                str("https://" + self.ceHost + ":8443/arex"),
                arc.Endpoint.JOBSUBMIT,
                "org.ogf.glue.emies.activitycreation",
            )

        # Submit jobs iteratively for now. Tentatively easier than mucking around with the JobSupervisor class
        for __i in range(numberOfJobs):
            # The basic job description
            jobdescs = arc.JobDescriptionList()
            # Get the job into the ARC way
            xrslString, diracStamp = self.__writeXRSL(executableFile)
            self.log.debug("XRSL string submitted : %s" % xrslString)
            self.log.debug("DIRAC stamp for job : %s" % diracStamp)
            # The arc bindings don't accept unicode objects in Python 2 so xrslString must be explicitly cast
            result = arc.JobDescription_Parse(str(xrslString), jobdescs)
            if not result:
                self.log.error("Invalid job description", "%r, message=%s" % (xrslString, result.str()))
                break
            # Submit the job
            jobs = arc.JobList()  # filled by the submit process
            submitter = arc.Submitter(self.usercfg)
            result = submitter.Submit(endpoint, jobdescs, jobs)
            # Save info or else ..else.
            if result == arc.SubmissionStatus.NONE:
                # Job successfully submitted
                pilotJobReference = jobs[0].JobID
                batchIDList.append(pilotJobReference)
                stampDict[pilotJobReference] = diracStamp
                self.log.debug("Successfully submitted job %s to CE %s" % (pilotJobReference, self.ceHost))
            else:
                message = "Failed to submit job because "
                if result.isSet(arc.SubmissionStatus.NOT_IMPLEMENTED):  # pylint: disable=no-member
                    self.log.warn("%s feature not implemented on CE? (weird I know - complain to site admins" % message)
                if result.isSet(arc.SubmissionStatus.NO_SERVICES):  # pylint: disable=no-member
                    self.log.warn("%s no services are running on CE? (open GGUS ticket to site admins" % message)
                if result.isSet(arc.SubmissionStatus.ENDPOINT_NOT_QUERIED):  # pylint: disable=no-member
                    self.log.warn("%s endpoint was not even queried. (network ..?)" % message)
                if result.isSet(arc.SubmissionStatus.BROKER_PLUGIN_NOT_LOADED):  # pylint: disable=no-member
                    self.log.warn("%s BROKER_PLUGIN_NOT_LOADED : ARC library installation problem?" % message)
                if result.isSet(arc.SubmissionStatus.DESCRIPTION_NOT_SUBMITTED):  # pylint: disable=no-member
                    self.log.warn(
                        "%s Job not submitted - incorrect job description? (missing field in XRSL string?)" % message
                    )
                if result.isSet(arc.SubmissionStatus.SUBMITTER_PLUGIN_NOT_LOADED):  # pylint: disable=no-member
                    self.log.warn("%s SUBMITTER_PLUGIN_NOT_LOADED : ARC library installation problem?" % message)
                if result.isSet(arc.SubmissionStatus.AUTHENTICATION_ERROR):  # pylint: disable=no-member
                    self.log.warn(
                        "%s authentication error - screwed up / expired proxy? Renew / upload pilot proxy on machine?"
                        % message
                    )
                if result.isSet(arc.SubmissionStatus.ERROR_FROM_ENDPOINT):  # pylint: disable=no-member
                    self.log.warn("%s some error from the CE - possibly CE problems?" % message)
                self.log.warn("%s ... maybe above messages will give a hint." % message)
                break  # Boo hoo *sniff*

        if batchIDList:
            result = S_OK(batchIDList)
            result["PilotStampDict"] = stampDict
        else:
            result = S_ERROR("No pilot references obtained from the ARC job submission")
        return result

    #############################################################################
    def killJob(self, jobIDList):
        """Kill the specified jobs"""

        result = self._prepareProxy()
        if not result["OK"]:
            self.log.error("ARCComputingElement: failed to set up proxy", result["Message"])
            return result
        self.usercfg.ProxyPath(os.environ["X509_USER_PROXY"])

        jobList = list(jobIDList)
        if isinstance(jobIDList, six.string_types):
            jobList = [jobIDList]

        self.log.debug("Killing jobs %s" % jobIDList)
        jobs = []
        for jobID in jobList:
            jobs.append(self.__getARCJob(jobID))

        # JobSupervisor is able to aggregate jobs to perform bulk operations and thus minimizes the communication overhead
        # We still need to create chunks to avoid timeout in the case there are too many jobs to supervise
        for chunk in breakListIntoChunks(jobs, 100):
            job_supervisor = arc.JobSupervisor(self.usercfg, chunk)
            if not job_supervisor.Cancel():
                errorString = " - ".join(jobList).strip()
                return S_ERROR("Failed to kill at least one of these jobs: %s. CE(?) not reachable?" % errorString)

        return S_OK()

    #############################################################################
    def getCEStatus(self):
        """Method to return information on running and pending jobs.
        We hope to satisfy both instances that use robot proxies and those which use proper configurations.
        """

        result = self._prepareProxy()
        if not result["OK"]:
            self.log.error("ARCComputingElement: failed to set up proxy", result["Message"])
            return result
        self.usercfg.ProxyPath(os.environ["X509_USER_PROXY"])

        # Try to find out which VO we are running for.
        vo = ""
        res = getVOfromProxyGroup()
        if res["OK"]:
            vo = res["Value"]

        result = S_OK()
        result["SubmittedJobs"] = 0
        if not vo:
            # Presumably the really proper way forward once the infosys-discuss WG comes up with a solution
            # and it is implemented. Needed for DIRAC instances which use robot certificates for pilots.
            if self.endpointType == "Gridftp":
                endpoints = [
                    arc.Endpoint(
                        str("ldap://" + self.ceHost + "/MDS-Vo-name=local,o=grid"),
                        arc.Endpoint.COMPUTINGINFO,
                        "org.nordugrid.ldapng",
                    )
                ]
            else:
                endpoints = [
                    arc.Endpoint(
                        str("https://" + self.ceHost + ":8443/arex"),
                        arc.Endpoint.COMPUTINGINFO,
                        "org.ogf.glue.emies.resourceinfo",
                    )
                ]

            retriever = arc.ComputingServiceRetriever(self.usercfg, endpoints)
            retriever.wait()  # Takes a bit of time to get and parse the ldap information
            targets = retriever.GetExecutionTargets()
            ceStats = targets[0].ComputingShare
            self.log.debug("Running jobs for CE %s : %s" % (self.ceHost, ceStats.RunningJobs))
            self.log.debug("Waiting jobs for CE %s : %s" % (self.ceHost, ceStats.WaitingJobs))
            result["RunningJobs"] = ceStats.RunningJobs
            result["WaitingJobs"] = ceStats.WaitingJobs
        else:
            # The system which works properly at present for ARC CEs that are configured correctly.
            # But for this we need the VO to be known - ask me (Raja) for the whole story if interested.
            # cmd = 'ldapsearch -x -LLL -H ldap://%s:2135 -b mds-vo-name=resource,o=grid "(GlueVOViewLocalID=%s)"' % (
            #     self.ceHost, vo.lower())
            if not self.queue:
                self.log.error("ARCComputingElement: No queue ...")
                res = S_ERROR("Unknown queue (%s) failure for site %s" % (self.queue, self.ceHost))
                return res
            cmd1 = "ldapsearch -x -o ldif-wrap=no -LLL -h %s:2135  -b 'o=glue' " % self.ceHost
            cmd2 = '"(&(objectClass=GLUE2MappingPolicy)(GLUE2PolicyRule=vo:%s))"' % vo.lower()
            cmd3 = " | grep GLUE2MappingPolicyShareForeignKey | grep %s" % (self.queue.split("-")[-1])
            cmd4 = " | sed 's/GLUE2MappingPolicyShareForeignKey: /GLUE2ShareID=/' "
            cmd5 = " | xargs -L1 ldapsearch -x -o ldif-wrap=no -LLL -h %s:2135 -b 'o=glue' " % self.ceHost
            cmd6 = " | egrep '(ShareWaiting|ShareRunning)'"
            res = shellCall(0, cmd1 + cmd2 + cmd3 + cmd4 + cmd5 + cmd6)
            if not res["OK"]:
                self.log.debug("Could not query CE %s - is it down?" % self.ceHost)
                return res
            try:
                ldapValues = res["Value"][1].split("\n")
                running = [lValue for lValue in ldapValues if "GLUE2ComputingShareRunningJobs" in lValue]
                waiting = [lValue for lValue in ldapValues if "GLUE2ComputingShareWaitingJobs" in lValue]
                result["RunningJobs"] = int(running[0].split(":")[1])
                result["WaitingJobs"] = int(waiting[0].split(":")[1])
            except IndexError:
                res = S_ERROR("Unknown ldap failure for site %s" % self.ceHost)
                return res

        return result

    #############################################################################
    def getJobStatus(self, jobIDList):
        """Get the status information for the given list of jobs"""

        result = self._prepareProxy()
        if not result["OK"]:
            self.log.error("ARCComputingElement: failed to set up proxy", result["Message"])
            return result
        self.usercfg.ProxyPath(os.environ["X509_USER_PROXY"])

        jobTmpList = list(jobIDList)
        if isinstance(jobIDList, six.string_types):
            jobTmpList = [jobIDList]

        # Pilots are stored with a DIRAC stamp (":::XXXXX") appended
        jobList = []
        for j in jobTmpList:
            if ":::" in j:
                job = j.split(":::")[0]
            else:
                job = j
            jobList.append(job)

        jobs = []
        for jobID in jobList:
            jobs.append(self.__getARCJob(jobID))

        # JobSupervisor is able to aggregate jobs to perform bulk operations and thus minimizes the communication overhead
        # We still need to create chunks to avoid timeout in the case there are too many jobs to supervise
        jobsUpdated = []
        for chunk in breakListIntoChunks(jobs, 100):
            job_supervisor = arc.JobSupervisor(self.usercfg, chunk)
            job_supervisor.Update()
            jobsUpdated.extend(job_supervisor.GetAllJobs())

        resultDict = {}
        jobsToRenew = []
        jobsToCancel = []
        for job in jobsUpdated:
            jobID = job.JobID
            self.log.debug("Retrieving status for job %s" % jobID)
            arcState = job.State.GetGeneralState()
            self.log.debug("ARC status for job %s is %s" % (jobID, arcState))
            if arcState:  # Meaning arcState is filled. Is this good python?
                resultDict[jobID] = self.mapStates[arcState]
                # Renew proxy only of jobs which are running or queuing
                if arcState in ("Running", "Queuing"):
                    nearExpiry = arc.Time() + arc.Period(10000)  # 2 hours, 46 minutes and 40 seconds
                    if job.ProxyExpirationTime < nearExpiry:
                        # Jobs to renew are aggregated to perform bulk operations
                        jobsToRenew.append(job)
                        self.log.debug(
                            "Renewing proxy for job %s whose proxy expires at %s" % (jobID, job.ProxyExpirationTime)
                        )
                if arcState == "Hold":
                    # Jobs to cancel are aggregated to perform bulk operations
                    # Cancel held jobs so they don't sit in the queue forever
                    jobsToCancel.append(job)
                    self.log.debug("Killing held job %s" % jobID)
            else:
                resultDict[jobID] = PilotStatus.UNKNOWN
            # If done - is it really done? Check the exit code
            if resultDict[jobID] == PilotStatus.DONE:
                exitCode = int(job.ExitCode)
                if exitCode:
                    resultDict[jobID] = PilotStatus.FAILED
            self.log.debug("DIRAC status for job %s is %s" % (jobID, resultDict[jobID]))

        # JobSupervisor is able to aggregate jobs to perform bulk operations and thus minimizes the communication overhead
        # We still need to create chunks to avoid timeout in the case there are too many jobs to supervise
        for chunk in breakListIntoChunks(jobsToRenew, 100):
            job_supervisor_renew = arc.JobSupervisor(self.usercfg, chunk)
            if not job_supervisor_renew.Renew():
                self.log.warn("At least one of the jobs failed to renew its credentials")

        for chunk in breakListIntoChunks(jobsToCancel, 100):
            job_supervisor_cancel = arc.JobSupervisor(self.usercfg, chunk)
            if not job_supervisor_cancel.Cancel():
                self.log.warn("At least one of the jobs failed to be cancelled")

        if not resultDict:
            return S_ERROR("No job statuses returned")

        return S_OK(resultDict)

    #############################################################################
    def getJobOutput(self, jobID, localDir=None):
        """Get the specified job standard output and error files. If the localDir is provided,
        the output is returned as file in this directory. Otherwise, the output is returned
        as strings.
        """
        result = self._prepareProxy()
        if not result["OK"]:
            self.log.error("ARCComputingElement: failed to set up proxy", result["Message"])
            return result
        self.usercfg.ProxyPath(os.environ["X509_USER_PROXY"])

        if jobID.find(":::") != -1:
            pilotRef, stamp = jobID.split(":::")
        else:
            pilotRef = jobID
            stamp = ""
        if not stamp:
            return S_ERROR("Pilot stamp not defined for %s" % pilotRef)

        job = self.__getARCJob(pilotRef)

        arcID = os.path.basename(pilotRef)
        self.log.debug("Retrieving pilot logs for %s" % pilotRef)
        if "WorkingDirectory" in self.ceParameters:
            workingDirectory = os.path.join(self.ceParameters["WorkingDirectory"], arcID)
        else:
            workingDirectory = arcID
        outFileName = os.path.join(workingDirectory, "%s.out" % stamp)
        errFileName = os.path.join(workingDirectory, "%s.err" % stamp)
        self.log.debug("Working directory for pilot output %s" % workingDirectory)

        isItOkay = job.Retrieve(self.usercfg, arc.URL(str(workingDirectory)), False)
        if isItOkay:
            output = None
            error = None
            try:
                with open(outFileName, "r") as outFile:
                    output = outFile.read()
                os.unlink(outFileName)
                with open(errFileName, "r") as errFile:
                    error = errFile.read()
                os.unlink(errFileName)
            except IOError as e:
                self.log.error("Error downloading outputs", repr(e).replace(",)", ")"))
                return S_ERROR("Error downloading outputs")
            self.log.debug("Pilot output = %s" % output)
            self.log.debug("Pilot error = %s" % error)
        else:
            job.Update()
            arcState = job.State.GetGeneralState()
            if arcState != "Undefined":
                return S_ERROR(
                    "Failed to retrieve output for %s as job is not finished (maybe not started yet)" % jobID
                )
            self.log.debug(
                "Could not retrieve pilot output for %s - either permission / proxy error or could not connect to CE"
                % pilotRef
            )
            return S_ERROR("Failed to retrieve output for %s" % jobID)

        return S_OK((output, error))



""" Test the SSLTransport mechanism """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import socket
import threading

try:
    import selectors
except ImportError:
    import selectors2 as selectors

from pytest import fixture

from diraccfg import CFG
from DIRAC.Core.Security.test.x509TestUtilities import CERTDIR, USERCERT, getCertOption
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.Core.DISET.private.Transports import PlainTransport, M2SSLTransport

# TODO: Expired hostcert
# TODO: Expired usercert
# TODO: Expired proxy
# TODO: Invalid/missing CA
# TODO: Connect Timeouts
# TODO: SSL Algorithms & Ciphers
# TODO: Missing hostcert
# TODO: Missing usercert
# TODO: Missing proxy
# TODO: Session test?
# TODO: Reload of CAs?

# Define all the locations

caLocation = os.path.join(CERTDIR, "ca")
hostCertLocation = os.path.join(CERTDIR, "host/hostcert.pem")
hostKeyLocation = os.path.join(CERTDIR, "host/hostkey.pem")
proxyFile = os.path.join(os.path.dirname(__file__), "proxy.pem")


MAGIC_QUESTION = "Who let the dog out"
MAGIC_ANSWER = "Who, Who, who ?"

PORT_NUMBER = 50000

# Transports are now tested in pairs:
# "Server-Client"
# Each pair is defined as a string.
TRANSPORTTESTS = ("Plain-Plain", "M2-M2")


# https://www.ibm.com/developerworks/linux/library/l-openssl/index.html
# http://www.herongyang.com/Cryptography/


class DummyServiceReactor(object):
    """This class behaves like a ServiceReactor, except that it exists after treating a single request"""

    def __init__(self, transportObject, port):
        """c'tor

        :param transportObject: type of TransportObject we will use
        :param port: port to listen to
        """
        self.__prepared = False
        self.port = port
        self.transportObject = transportObject

        # Server transport object
        self.transport = None
        # Client connection
        self.clientTransport = None
        # Message received from the client
        self.receivedMessage = None

    def handleConnection(self, clientTransport):
        """This is normally done is Service.py in different thread
        It more or less does Service._processInThread
        """

        self.clientTransport = clientTransport
        res = clientTransport.handshake()
        assert res["OK"], res

        self.receivedMessage = clientTransport.receiveData(1024)
        clientTransport.sendData(MAGIC_ANSWER)
        clientTransport.close()

    def prepare(self):
        """Start listening"""
        if not self.__prepared:
            self.__createListeners()
        self.__prepared = True

    def serve(self):
        """Wait for connections and handle the first one."""
        self.prepare()
        self.__acceptIncomingConnection()

    def __createListeners(self):
        """Create the listener transport"""
        self.transport = self.transportObject(("", self.port), bServerMode=True)
        res = self.transport.initAsServer()
        assert res["OK"]

    def __acceptIncomingConnection(self):
        """
        This method just gets the incoming connection, and handle it, once.
        """
        try:
            sel = selectors.DefaultSelector()
            sel.register(self.transport.getSocket(), selectors.EVENT_READ)
            assert sel.select(timeout=2)
            result = self.transport.acceptConnection()
            assert result["OK"], result
            clientTransport = result["Value"]

            self.handleConnection(clientTransport)

        except socket.error:
            return

    def closeListeningConnections(self):
        """Close the connection"""
        self.transport.close()


def transportByName(transport):
    """helper function to get a transport class by 'friendly' name."""
    if transport.lower() == "plain":
        return PlainTransport.PlainTransport
    elif transport.lower() == "m2":
        return M2SSLTransport.SSLTransport
    raise RuntimeError("Unknown Transport Name: %s" % transport)


@fixture(scope="function", params=TRANSPORTTESTS)
def create_serverAndClient(request):
    """This function starts a server, and closes it after
    The server will use the parametrized transport type
    """

    # Reinitialize the configuration.
    # We do it here rather than at the start of the module
    # to accommodate for pytest when going through all the DIRAC tests

    gConfigurationData.localCFG = CFG()
    gConfigurationData.remoteCFG = CFG()
    gConfigurationData.mergedCFG = CFG()
    gConfigurationData.generateNewVersion()
    gConfigurationData.setOptionInCFG("/DIRAC/Security/CALocation", caLocation)
    gConfigurationData.setOptionInCFG("/DIRAC/Security/CertFile", hostCertLocation)
    gConfigurationData.setOptionInCFG("/DIRAC/Security/KeyFile", hostKeyLocation)

    testStr = request.param
    serverName, clientName = testStr.split("-")
    serverClass = transportByName(serverName)
    clientClass = transportByName(clientName)

    sr = DummyServiceReactor(serverClass, PORT_NUMBER)
    server_thread = threading.Thread(target=sr.serve)
    sr.prepare()
    server_thread.start()

    # Create the client
    clientOptions = {
        "clientMode": True,
        "proxyLocation": proxyFile,
    }
    clientTransport = clientClass(("localhost", PORT_NUMBER), bServerMode=False, **clientOptions)
    res = clientTransport.initAsClient()
    assert res["OK"], res

    yield sr, clientTransport

    clientTransport.close()
    sr.closeListeningConnections()
    server_thread.join()

    # Clean the config
    gConfigurationData.localCFG = CFG()
    gConfigurationData.remoteCFG = CFG()
    gConfigurationData.mergedCFG = CFG()
    gConfigurationData.generateNewVersion()


def ping_server(clientTransport):
    """This sends a message to the server and expects an answer
    This basically does the same as BaseClient.py

    :param clientTransport: the Transport object to be used as client
    """

    clientTransport.setSocketTimeout(5)
    result = clientTransport.sendData(MAGIC_QUESTION)
    assert result["OK"]
    serverReturn = clientTransport.receiveData()
    return serverReturn


def test_simpleMessage(create_serverAndClient):
    """Send a message, wait for an answer"""
    serv, client = create_serverAndClient
    serverAnswer = ping_server(client)
    assert serv.receivedMessage == MAGIC_QUESTION
    assert serverAnswer == MAGIC_ANSWER


def test_getRemoteInfo(create_serverAndClient):
    """Check the information from remote peer"""
    serv, client = create_serverAndClient
    ping_server(client)

    addr_info = client.getRemoteAddress()
    assert addr_info[0] in ("127.0.0.1", "::ffff:127.0.0.1", "::1")
    assert addr_info[1] == PORT_NUMBER
    # The peer credentials are not filled on the client side
    assert client.peerCredentials == {}

    # We do not know about the port, so check only the address, taking into account bloody IPv6
    assert serv.clientTransport.getRemoteAddress()[0] in ("127.0.0.1", "::ffff:127.0.0.1", "::1")
    peerCreds = serv.clientTransport.peerCredentials

    # There are no credentials for PlainTransport
    if client.__class__.__name__ == "PlainTransport":
        assert peerCreds == {}
    else:
        assert peerCreds["DN"] == getCertOption(USERCERT, "subjectDN")
        assert peerCreds["x509Chain"].getNumCertsInChain()["Value"] == 2
        assert peerCreds["isProxy"] is True
        assert peerCreds["isLimitedProxy"] is False



"""
   Collection of DIRAC useful operating system related modules
   by default on Error they return None
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import six
import os
import distutils.spawn  # pylint: disable=no-name-in-module,import-error

import DIRAC
from DIRAC.Core.Utilities.Subprocess import shellCall, systemCall
from DIRAC.Core.Utilities import List

__RCSID__ = "$Id$"


DEBUG = 0


def uniquePath(path=None):
    """
    Utility to squeeze the string containing a PATH-like value to
    leave only unique elements preserving the original order
    """
    if not isinstance(path, six.string_types):
        return None

    try:
        elements = List.uniqueElements(List.fromChar(path, ":"))
        return ":".join(elements)
    except Exception:
        return None


def getDiskSpace(path=".", exclude=None):
    """Get the free disk space in the partition containing the path.
    The disk space is reported in MBytes. Returned 0 in case of any
    error, e.g. path does not exist
    """

    if not os.path.exists(path):
        return -1
    comm = "df -P -m %s " % path
    if exclude:
        comm += "-x %s " % exclude
    comm += "| tail -1"
    resultDF = shellCall(10, comm)
    if resultDF["OK"] and not resultDF["Value"][0]:
        output = resultDF["Value"][1]
        if output.find(" /afs") >= 0:  # AFS disk space
            comm = "fs lq | tail -1"
            resultAFS = shellCall(10, comm)
            if resultAFS["OK"] and not resultAFS["Value"][0]:
                output = resultAFS["Value"][1]
                fields = output.split()
                quota = int(fields[1])
                used = int(fields[2])
                space = (quota - used) / 1024
                return int(space)
            else:
                return -1
        else:
            fields = output.split()
            try:
                value = int(fields[3])
            except Exception as error:
                print("Exception during disk space evaluation:", str(error))
                value = -1
            return value
    else:
        return -1


def getDirectorySize(path):
    """Get the total size of the given directory in MB"""

    comm = "du -s -m %s" % path
    result = shellCall(10, comm)
    if not result["OK"] or result["Value"][0] != 0:
        return 0
    else:
        output = result["Value"][1]
        print(output)
        size = int(output.split()[0])
        return size


def sourceEnv(timeout, cmdTuple, inputEnv=None):
    """Function to source configuration files in a platform dependent way and get
    back the environment
    """

    # add appropriate extension to first element of the tuple (the command)
    envAsDict = '&& python -c "import os,sys ; print >> sys.stderr, os.environ"'

    # 1.- Choose the right version of the configuration file
    if DIRAC.getPlatformTuple()[0] == "Windows":
        cmdTuple[0] += ".bat"
    else:
        cmdTuple[0] += ".sh"

    # 2.- Check that it exists
    if not os.path.exists(cmdTuple[0]):
        result = DIRAC.S_ERROR("Missing script: %s" % cmdTuple[0])
        result["stdout"] = ""
        result["stderr"] = "Missing script: %s" % cmdTuple[0]
        return result

    # Source it in a platform dependent way:
    # On windows the execution makes the environment to be inherit
    # On Linux or Darwin use bash and source the file.
    if DIRAC.getPlatformTuple()[0] == "Windows":
        # this needs to be tested
        cmd = " ".join(cmdTuple) + envAsDict
        ret = shellCall(timeout, [cmd], env=inputEnv)
    else:
        cmdTuple.insert(0, "source")
        cmd = " ".join(cmdTuple) + envAsDict
        ret = systemCall(timeout, ["/bin/bash", "-c", cmd], env=inputEnv)

    # 3.- Now get back the result
    stdout = ""
    stderr = ""
    result = DIRAC.S_OK()
    if ret["OK"]:
        # The Command has not timeout, retrieve stdout and stderr
        stdout = ret["Value"][1]
        stderr = ret["Value"][2]
        if ret["Value"][0] == 0:
            # execution was OK
            try:
                result["outputEnv"] = eval(stderr.split("\n")[-2] + "\n")
                stderr = "\n".join(stderr.split("\n")[:-2])
            except Exception:
                stdout = cmd + "\n" + stdout
                result = DIRAC.S_ERROR("Could not parse Environment dictionary from stderr")
        else:
            # execution error
            stdout = cmd + "\n" + stdout
            result = DIRAC.S_ERROR("Execution returns %s" % ret["Value"][0])
    else:
        # Timeout
        stdout = cmd
        stderr = ret["Message"]
        result = DIRAC.S_ERROR(stderr)

    # 4.- Put stdout and stderr in result structure
    result["stdout"] = stdout
    result["stderr"] = stderr

    return result


def which(executable):

    return distutils.spawn.find_executable(executable)  # pylint: disable=no-member



#!/usr/bin/env python
"""
Remove the given file or a list of files from the File Catalog and from the storage

Example:
  $ dirac-dms-remove-files /formation/user/v/vhamar/Test.txt
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Names"))
    Script.registerArgument(["LFN:       Logical File Names"], mandatory=False)
    Script.parseCommandLine()

    import os
    import DIRAC
    from DIRAC import gLogger

    first, lfns = Script.getPositionalArgs(group=True)
    if os.path.exists(first):
        with open(first, "r") as inputFile:
            string = inputFile.read()
        lfns.extend([lfn.strip() for lfn in string.splitlines()])
    else:
        lfns.insert(0, first)

    from DIRAC.Core.Utilities.List import breakListIntoChunks
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager

    dm = DataManager()

    errorReasons = {}
    successfullyRemoved = 0
    for lfnList in breakListIntoChunks(lfns, 100):
        res = dm.removeFile(lfnList)
        if not res["OK"]:
            gLogger.error("Failed to remove data", res["Message"])
            DIRAC.exit(-2)
        for lfn, r in res["Value"]["Failed"].items():
            reason = str(r)
            if reason not in errorReasons:
                errorReasons[reason] = []
            errorReasons[reason].append(lfn)
        successfullyRemoved += len(res["Value"]["Successful"])

    for reason, lfns in errorReasons.items():
        gLogger.notice("Failed to remove %d files with error: %s" % (len(lfns), reason))
    if successfullyRemoved > 0:
        gLogger.notice("Successfully removed %d files" % successfullyRemoved)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Do the initial installation and configuration of a DIRAC component
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import exit as DIRACexit
from DIRAC import gConfig, gLogger, S_OK
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities.Extensions import extensionsByPriority
from DIRAC.FrameworkSystem.Utilities import MonitoringUtilities

__RCSID__ = "$Id$"

overwrite = False
module = ""
specialOptions = {}


def setOverwrite(opVal):
    global overwrite
    overwrite = True
    return S_OK()


def setModule(optVal):
    global specialOptions, module
    specialOptions["Module"] = optVal
    module = optVal
    return S_OK()


def setSpecialOption(optVal):
    global specialOptions
    option, value = optVal.split("=")
    specialOptions[option] = value
    return S_OK()


@Script()
def main():
    global overwrite
    global specialOptions
    global module
    global specialOptions

    from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller

    gComponentInstaller.exitOnError = True

    Script.registerSwitch("w", "overwrite", "Overwrite the configuration in the global CS", setOverwrite)
    Script.registerSwitch("m:", "module=", "Python module name for the component code", setModule)
    Script.registerSwitch("p:", "parameter=", "Special component option ", setSpecialOption)
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        (
            "System/Component: Full component name (ie: WorkloadManagement/Matcher)",
            "System:           Name of the DIRAC system (ie: WorkloadManagement)",
        )
    )
    Script.registerArgument(" Component:        Name of the DIRAC service (ie: Matcher)", mandatory=False)

    Script.parseCommandLine()
    args = Script.getPositionalArgs()

    if len(args) == 1:
        args = args[0].split("/")

    if len(args) != 2:
        Script.showHelp(exitCode=1)

    system = args[0]
    component = args[1]
    compOrMod = module or component

    result = gComponentInstaller.getSoftwareComponents(extensionsByPriority())
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    availableComponents = result["Value"]

    for compType in availableComponents:
        if system in availableComponents[compType] and compOrMod in availableComponents[compType][system]:
            cType = compType[:-1].lower()
            break
    else:
        gLogger.error("Component %s/%s is not available for installation" % (system, component))
        DIRACexit(1)

    if module:
        result = gComponentInstaller.addDefaultOptionsToCS(
            gConfig, cType, system, module, extensionsByPriority(), overwrite=overwrite
        )
        result = gComponentInstaller.addDefaultOptionsToCS(
            gConfig,
            cType,
            system,
            component,
            extensionsByPriority(),
            specialOptions=specialOptions,
            overwrite=overwrite,
            addDefaultOptions=False,
        )
    else:
        result = gComponentInstaller.addDefaultOptionsToCS(
            gConfig,
            cType,
            system,
            component,
            extensionsByPriority(),
            specialOptions=specialOptions,
            overwrite=overwrite,
        )

    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    result = gComponentInstaller.installComponent(cType, system, component, extensionsByPriority(), module)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    gLogger.notice("Successfully installed component %s in %s system, now setting it up" % (component, system))
    result = gComponentInstaller.setupComponent(cType, system, component, extensionsByPriority(), module)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    if component == "ComponentMonitoring":
        result = MonitoringUtilities.monitorInstallation("DB", system, "InstalledComponentsDB")
        if not result["OK"]:
            gLogger.error(result["Message"])
            DIRACexit(1)
    result = MonitoringUtilities.monitorInstallation(cType, system, component, module)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    gLogger.notice("Successfully completed the installation of %s/%s" % (system, component))
    DIRACexit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :    dirac-framework-ping-service
# Author :  Stuart Paterson
########################################################################
"""
Ping the given DIRAC Service

Example:
  $ dirac-framework-ping-service WorkloadManagement PilotManager
  {'OK': True,
   'Value': {'cpu times': {'children system time': 0.0,
                           'children user time': 0.0,
                           'elapsed real time': 8778481.7200000007,
                           'system time': 54.859999999999999,
                           'user time': 361.06999999999999},
             'host uptime': 4485212L,
             'load': '3.44 3.90 4.02',
             'name': 'WorkloadManagement/PilotManager',
             'service start time': datetime.datetime(2011, 2, 21, 8, 58, 35, 521438),
             'service uptime': 85744,
             'service url': 'dips://dirac.in2p3.fr:9171/WorkloadManagement/PilotManager',
             'time': datetime.datetime(2011, 3, 14, 11, 47, 40, 394957),
             'version': 'v5r12-pre9'},
   'rpcStub': (('WorkloadManagement/PilotManager',
                {'delegatedDN': '/O=GRID-FR/C=FR/O=CNRS/OU=CPPM/CN=Vanessa Hamar',
                 'delegatedGroup': 'dirac_user',
                 'skipCACheck': True,
                 'timeout': 120}),
               'ping',
               ())}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        (
            "URL:            URL of the service to ping (instead of System and Service)",
            "System/Service: Full component name (ie: WorkloadManagement/Matcher)",
            "System:         Name of the DIRAC system (ie: WorkloadManagement)",
        )
    )
    Script.registerArgument(" Service:        Name of the DIRAC service (ie: Matcher)", mandatory=False)
    _, args = Script.parseCommandLine(ignoreErrors=True)
    system = None
    service = None
    url = None
    if len(args) == 1:
        # it is a URL
        if args[0].startswith("dips://"):
            url = args[0]
        # It is System/Service
        else:
            sys_serv = args[0].split("/")
            if len(sys_serv) != 2:
                Script.showHelp(exitCode=1)
            else:
                system, service = sys_serv

    elif len(args) == 2:
        system, service = args[0], args[1]
    else:
        Script.showHelp(exitCode=1)

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    exitCode = 0

    result = dirac.pingService(system, service, printOutput=True, url=url)

    if not result:
        print("ERROR: Null result from ping()")
        exitCode = 2
    elif not result["OK"]:
        print("ERROR: ", result["Message"])
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Container for TaskManager plug-ins, to handle the destination of the tasks
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC import gLogger

from DIRAC.Core.Utilities.List import fromChar
from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites
from DIRAC.TransformationSystem.Client.PluginBase import PluginBase

__RCSID__ = "$Id$"


class TaskManagerPlugin(PluginBase):
    """A TaskManagerPlugin object should be instantiated by every TaskManager object.

    self.params here could be
    {'Status': 'Created', 'TargetSE': 'Unknown', 'TransformationID': 1086L, 'RunNumber': 0L,
    'Site': 'DIRAC.Test.ch', 'TaskID': 21L, 'InputData': '', 'JobType': 'MCSimulation'}
    which corresponds to paramsDict in TaskManager (which is in fact a tasks dict)
    """

    def _BySE(self):
        """Matches using TargetSE. This is the standard plugin."""

        destSites = set()

        try:
            seList = ["Unknown"]
            if self.params["TargetSE"]:
                if isinstance(self.params["TargetSE"], six.string_types):
                    seList = fromChar(self.params["TargetSE"])
                elif isinstance(self.params["TargetSE"], list):
                    seList = self.params["TargetSE"]
        except KeyError:
            pass

        if not seList or seList == ["Unknown"]:
            return destSites

        for se in seList:
            res = getSitesForSE(se)
            if not res["OK"]:
                gLogger.warn("Could not get Sites associated to SE", res["Message"])
            else:
                thisSESites = res["Value"]
                if thisSESites:
                    # We make an OR of the possible sites
                    destSites.update(thisSESites)

        gLogger.debug("Destinations: %s" % ",".join(destSites))
        return destSites

    def _ByJobType(self):
        """By default, all sites are allowed to do every job.
        The actual rules are freely specified in the Operation JobTypeMapping section.
        The content of the section may look like this:

        User
        {
          Exclude = PAK
          Exclude += Ferrara
          Exclude += Bologna
          Exclude += Paris
          Exclude += CERN
          Exclude += IN2P3
          Allow
          {
            Paris = IN2P3
            CERN = CERN
            IN2P3 = IN2P3
          }
        }
        DataReconstruction
        {
          Exclude = PAK
          Exclude += Ferrara
          Exclude += CERN
          Exclude += IN2P3
          Allow
          {
            Ferrara = CERN
            CERN = CERN
            IN2P3 = IN2P3
            IN2P3 += CERN
          }
        }
        Merge
        {
          Exclude = ALL
          Allow
          {
            CERN = CERN
            IN2P3 = IN2P3
          }
        }

        The sites in the exclusion list will be removed.
        The allow section says where each site may help another site

        """
        # 1. get sites list
        res = getSites()
        if not res["OK"]:
            gLogger.error("Could not get the list of sites", res["Message"])
            return res
        destSites = set(res["Value"])

        # 2. get JobTypeMapping "Exclude" value (and add autoAddedSites)
        gLogger.debug("Getting JobTypeMapping 'Exclude' value (and add autoAddedSites)")
        jobType = self.params["JobType"]
        if not jobType:
            raise RuntimeError("No jobType specified")
        excludedSites = set(self.opsH.getValue("JobTypeMapping/%s/Exclude" % jobType, []))
        gLogger.debug("Explicitly excluded sites for %s task: %s" % (jobType, ",".join(excludedSites)))
        autoAddedSites = self.opsH.getValue("JobTypeMapping/AutoAddedSites", [])
        if "WithStorage" in autoAddedSites:
            # Add all sites with storage, such that jobs can run wherever data is
            autoAddedSites.remove("WithStorage")
            autoAddedSites += DMSHelpers().getTiers(withStorage=True)

        # 3. removing sites in Exclude
        if not excludedSites:
            pass
        elif "ALL" in excludedSites:
            destSites = set()
        else:
            destSites -= excludedSites

        # 4. get JobTypeMapping "Allow" section
        res = self.opsH.getOptionsDict("JobTypeMapping/%s/Allow" % jobType)
        if not res["OK"]:
            gLogger.debug(res["Message"])
            allowed = {}
        else:
            allowed = dict((site, set(fromChar(fromSites))) for site, fromSites in res["Value"].items())

        autoAddedSites = set(self.opsH.getValue("JobTypeMapping/%s/AutoAddedSites" % jobType, autoAddedSites))
        gLogger.debug("Auto-added sites for %s task: %s" % (jobType, ",".join(autoAddedSites)))
        # 5. add autoAddedSites, if requested
        for autoAddedSite in autoAddedSites:
            allowed.setdefault(autoAddedSite, set()).add(autoAddedSite)
        gLogger.debug("Allowed sites for %s task: %s" % (jobType, ",".join(allowed)))

        # 6. Allowing sites that should be allowed
        taskSiteDestination = self._BySE()

        for destSite, fromSites in allowed.items():
            for fromSite in fromSites:
                if not taskSiteDestination or fromSite in taskSiteDestination:
                    destSites.add(destSite)

        gLogger.debug(
            "Computed list of destination sites for %s task with TargetSE %s: %s"
            % (jobType, self.params["TargetSE"], ",".join(destSites))
        )
        return destSites



"""Functions that assert conditions for use in pytest tests.

The AgentOptionsTest can be used to check consistency of options for Agents.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import inspect
import importlib
import logging
import os

from pprint import pformat

from mock import patch, call, MagicMock as Mock
from diraccfg import CFG

import DIRAC

LOG = logging.getLogger(__name__)


def _parseOption(outDict, inDict, optionPrefix=""):
    """Parse the ConfigTemplates options.

    Handle some special cases.
    """
    LOG.info("Parsing into %s, from %s, prefix %r", outDict, inDict, optionPrefix)
    for option, value in inDict.items():
        optionName = "/".join([optionPrefix, option]).strip("/")
        LOG.info("Parsing %r with %r", optionName, value)
        if isinstance(value, six.string_types) and value.lower() in ("no", "false"):
            outDict[optionName] = False
        elif isinstance(value, six.string_types) and value.lower() in ("yes", "true"):
            outDict[optionName] = True
        elif isinstance(value, six.string_types) and "," in value:
            outDict[optionName] = [val.strip() for val in value.split(",")]
        elif isinstance(value, dict):
            _parseOption(outDict, value, optionPrefix=optionName)
        elif isinstance(value, six.string_types):
            outDict[optionName] = value
            if value.isdigit():
                try:
                    outDict[optionName] = int(value)
                except ValueError:
                    pass
            else:
                try:
                    outDict[optionName] = float(value)
                except ValueError:
                    pass


def AgentOptionsTest(agentPath, options, mocker):
    """Test the consistency of options in ConfigTemplate and initialize of the agent.

    :param str agentPath: Module where the agent can be found, e.g. DIRAC.Core.Agent.CoreAgent
    :param list ignoreOptions: list of options to ignore during checks
    :param mocker: the mocker fixture from pytest-mock
    """
    agentPathSplit = agentPath.split(".")
    systemName = agentPathSplit[1]
    agentName = agentPathSplit[-1]

    agentModule = importlib.import_module(agentPath)
    LOG.info("Agents: %s %s", agentPath, agentModule)
    agentClass = None
    agentLocation = os.path.dirname(agentModule.__file__)

    options = options if options is not None else {}
    ignoreOptions = options.get("IgnoreOptions", [])
    specialMocks = options.get("SpecialMocks", {})

    # mock everything but the agentClass
    for name, member in inspect.getmembers(agentModule):
        LOG.info("Mocking? %s, %s, %s, isclass(%s)", name, callable(member), type(member), inspect.isclass(member))
        if name != "AgentModule" and "_AgentModule__executeModuleCycle" in dir(member):
            LOG.info("Found the agent class %s, %s", name, member)
            agentClass = member
            continue
        elif name == "AgentModule":
            continue
        if callable(member) or inspect.ismodule(member):
            LOG.info("Mocking: %s, %s, %s", name, member, type(member))
            mocker.patch(agentPath + "." + name, new=Mock())

    if specialMocks is not None:
        for name, retVal in specialMocks.items():
            mocker.patch(agentPath + "." + name, new=Mock(return_value=retVal))

    def returnDefault(*args):
        if len(args) > 1:
            LOG.debug("ReturningDefault: %s, %s", args, type(args[1]))
            return args[1]
        LOG.debug("ReturningDefault: None")
        return None

    getOptionMock = Mock(name="am_getOption", side_effect=returnDefault)

    def instrument(*args, **kwargs):
        """Mock some functions that come from the AgentModule and are not present otherwise."""
        args[0].am_getControlDirectory = Mock()
        args[0].am_getOption = getOptionMock
        args[0].am_getModuleParam = Mock()
        args[0].am_getWorkDirectory = Mock()
        args[0].am_setOption = Mock()
        args[0].log = Mock()
        return None

    initMock = Mock(side_effect=instrument)

    class MockAgentModule(object):
        def __init__(self, *args, **kwargs):
            instrument(self)

    patchBase = patch.object(agentClass, "__bases__", (MockAgentModule,))
    with patchBase, patch(agentPath + ".AgentModule.__init__", new=initMock), patch(
        "DIRAC.Core.Base.AgentModule.AgentModule.am_getOption", new=getOptionMock
    ):
        patchBase.is_local = True
        agentInstance = agentClass(agentName="sys/name", loadName="sys/name")
        instrument(agentInstance)

    for func in ["initialize", "beginExecution"]:
        if hasattr(agentInstance, func):
            getattr(agentInstance, func)()
    checkAgentOptions(getOptionMock, systemName, agentName, agentLocation, ignoreOptions=ignoreOptions)


def checkAgentOptions(getOptionMock, systemName, agentName, agentLocation, ignoreOptions=None):
    """Ensure that all the agent options are properly documented.

    :param getOptionMock: Mock object for agentmodule.get_amOption function
    :param str systemName: name of the **System**
    :param str agentName: name of the **Agent**
    :param list ignoreOptions: list of options to ignore
    """
    if ignoreOptions is None:
        ignoreOptions = []

    # add some options that can be set, see the AgentModule for all of them
    ignoreOptions.extend(
        ["PollingTime", "Status", "Enabled", "MaxCycles", "LogOutputs", "ControlDirectory", "shifterProxy"]
    )
    ignoreOptions = list(set(ignoreOptions))
    config = CFG()

    LOG.info("Testing %s/%s, ignoring options %s", systemName, agentName, ignoreOptions)

    # expect the ConfigTemplate one level above the agent module
    configFilePath = os.path.join(agentLocation, "..", "ConfigTemplate.cfg")
    config.loadFromFile(configFilePath)
    optionsDict = config.getAsDict("Agents/%s" % agentName)
    outDict = {}
    _parseOption(outDict, optionsDict)
    optionsDict = outDict
    LOG.info("Calls: %s", pformat(getOptionMock.call_args_list))
    LOG.info("Options found in ConfigTemplate: %s ", list(optionsDict.keys()))

    # check that values in ConfigTemplate are used
    for option, value in optionsDict.items():
        if any(ignoreOp in option for ignoreOp in ignoreOptions):
            LOG.info("From Agent: ignoring option %r with value %r, (%s)", option, value, type(value))
            continue
        LOG.info("Looking for call to option %r with value %r, (%s)", option, value, type(value))
        if not isinstance(value, bool) and not value:  # empty string, list, dict ...
            assert any(call(option, null) in getOptionMock.call_args_list for null in ({}, set(), [], "", 0, None))
        else:
            assert (
                call(option, value) in getOptionMock.call_args_list
                or call(option, [value]) in getOptionMock.call_args_list
            )

    # check that options used in the agent are in the ConfigTemplates
    for opCall in getOptionMock.call_args_list:
        optionArguments = opCall[0]
        if len(optionArguments) != 2:
            continue
        optionName = optionArguments[0]
        optionValue = optionArguments[1]
        if optionName in ignoreOptions:
            LOG.info("From Template: ignoring option %r with %r", optionName, optionValue)
            continue
        LOG.info("Checking Template option %r with %r", optionName, optionValue)
        assert optionName in optionsDict
        if not optionsDict[optionName]:
            assert not optionValue
            continue
        assert optionsDict[optionName] == optionValue or [optionsDict[optionName]] == optionValue



########################################################################
# File :   ProxyProviderFactory.py
# Author : A.T.
########################################################################

"""  The Proxy Provider Factory instantiates ProxyProvider objects
     according to their configuration
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getInfoAboutProviders

__RCSID__ = "$Id$"


class ProxyProviderFactory(object):

    #############################################################################
    def __init__(self):
        """Standard constructor"""
        self.log = gLogger.getSubLogger(__name__)

    #############################################################################
    def getProxyProvider(self, proxyProvider):
        """This method returns a ProxyProvider instance corresponding to the supplied
        name.

        :param str proxyProvider: the name of the Proxy Provider

        :return: S_OK(ProxyProvider)/S_ERROR()
        """
        if not proxyProvider:
            return S_ERROR("Provider name not set.")
        result = getInfoAboutProviders(of="Proxy", providerName=proxyProvider, option="all", section="all")
        if not result["OK"]:
            return result
        ppDict = result["Value"]
        ppDict["ProviderName"] = proxyProvider
        ppType = ppDict.get("ProviderType")
        if not ppType:
            return S_ERROR("Cannot find information for ProxyProvider %s" % proxyProvider)
        self.log.verbose("Creating ProxyProvider of %s type with the name %s" % (ppType, proxyProvider))
        subClassName = "%sProxyProvider" % (ppType)

        result = ObjectLoader().loadObject("Resources.ProxyProvider.%s" % subClassName)
        if not result["OK"]:
            self.log.error("Failed to load object", "%s: %s" % (subClassName, result["Message"]))
            return result

        ppClass = result["Value"]
        try:
            pProvider = ppClass()
            pProvider.setParameters(ppDict)
        except Exception as x:
            msg = "ProxyProviderFactory could not instantiate %s object: %s" % (subClassName, str(x))
            self.log.exception()
            self.log.warn(msg)
            return S_ERROR(msg)

        return S_OK(pProvider)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-allow-site
# Author :  Stuart Paterson
########################################################################
"""
Add Site to Active mask for current Setup

Example:
  $ dirac-admin-allow-site LCG.IN2P3.fr "FRANCE"
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import time

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("E:", "email=", "Boolean True/False (True by default)")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("Site:     Name of the Site")
    Script.registerArgument("Comment:  Reason of the action")
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
    from DIRAC import exit as DIRACExit, gConfig, gLogger

    def getBoolean(value):
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        else:
            Script.showHelp()

    email = True
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "email":
            email = getBoolean(switch[1])

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []
    setup = gConfig.getValue("/DIRAC/Setup", "")
    if not setup:
        print("ERROR: Could not contact Configuration Service")
        exitCode = 2
        DIRACExit(exitCode)

    # result = promptUser(
    #     'All the elements that are associated with this site will be active, '
    #     'are you sure about this action?'
    # )
    # if not result['OK'] or result['Value'] is 'n':
    #  print 'Script stopped'
    #  DIRACExit( 0 )

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    site, comment = Script.getPositionalArgs(group=True)
    result = diracAdmin.allowSite(site, comment, printOutput=True)
    if not result["OK"]:
        errorList.append((site, result["Message"]))
        exitCode = 2
    else:
        if email:
            userName = diracAdmin._getCurrentUser()
            if not userName["OK"]:
                print("ERROR: Could not obtain current username from proxy")
                exitCode = 2
                DIRACExit(exitCode)
            userName = userName["Value"]
            subject = "%s is added in site mask for %s setup" % (site, setup)
            body = "Site %s is added to the site mask for %s setup by %s on %s.\n\n" % (
                site,
                setup,
                userName,
                time.asctime(),
            )
            body += "Comment:\n%s" % comment
            addressPath = "EMail/Production"
            address = Operations().getValue(addressPath, "")
            if not address:
                gLogger.notice("'%s' not defined in Operations, can not send Mail\n" % addressPath, body)
            else:
                result = diracAdmin.sendMail(address, subject, body)
        else:
            print("Automatic email disabled by flag.")

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



"""
   DIRAC.Interfaces.API package

   DIRAC - Distributed Infrastructure With Remote Agent Control

   DIRAC is a system which allows to integrate various distributed
   computing resources in a single system which is used for MC data
   production, data reprocessing and final user analysis. It consists
   of the following main subsystems:

     - Workload Management System
     - Data Management System
     - Information Services
     - Tools and utilities

   The DIRAC API provides a transparent and secure way for users to submit
   jobs to the Grid.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"



""" Cache for the Plotting service plots
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os.path
import time
import threading

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.Core.Utilities.Graphs import graph


class PlotCache(object):
    def __init__(self, plotsLocation=False):
        self.plotsLocation = plotsLocation
        self.alive = True
        self.__graphCache = DictCache(deleteFunction=_deleteGraph)
        self.__graphLifeTime = 600
        self.purgeThread = threading.Thread(target=self.purgeExpired)
        self.purgeThread.start()

    def setPlotsLocation(self, plotsDir):
        self.plotsLocation = plotsDir
        for plot in os.listdir(self.plotsLocation):
            if plot.find(".png") > 0:
                plotLocation = "%s/%s" % (self.plotsLocation, plot)
                gLogger.verbose("Purging %s" % plotLocation)
                os.unlink(plotLocation)

    def purgeExpired(self):
        while self.alive:
            time.sleep(self.__graphLifeTime)
            self.__graphCache.purgeExpired()

    def getPlot(self, plotHash, plotData, plotMetadata, subplotMetadata):
        """
        Get plot from the cache if exists, else generate it
        """

        plotDict = self.__graphCache.get(plotHash)
        if plotDict is None:
            basePlotFileName = "%s/%s.png" % (self.plotsLocation, plotHash)
            if subplotMetadata:
                retVal = graph(plotData, basePlotFileName, plotMetadata, metadata=subplotMetadata)
            else:
                retVal = graph(plotData, basePlotFileName, plotMetadata)
            if not retVal["OK"]:
                return retVal
            plotDict = retVal["Value"]
            if plotDict["plot"]:
                plotDict["plot"] = os.path.basename(basePlotFileName)
            self.__graphCache.add(plotHash, self.__graphLifeTime, plotDict)
        return S_OK(plotDict)

    def getPlotData(self, plotFileName):
        filename = "%s/%s" % (self.plotsLocation, plotFileName)
        try:
            with open(filename, "rb") as fd:
                data = fd.read()
        except Exception as v:
            return S_ERROR("Can't open file %s: %s" % (plotFileName, str(v)))
        return S_OK(data)


def _deleteGraph(plotDict):
    try:
        for key in plotDict:
            value = plotDict[key]
            if value and os.path.isfile(value):
                os.unlink(value)
    except Exception:
        pass


gPlotCache = PlotCache()



""" SummarizeLogsAgent module

  This agents scans all the log tables (SiteLog, ResourceLog and NodeLog) on the
  ResourceStatusDB and summarizes them. The results are stored on the History
  tables (SiteHistory, ResourceHistory and NodeHistory) and the Log tables
  cleared.

  In order to summarize the logs, all entries with no changes on the Status or
  TokenOwner column for a given (Name, StatusType) tuple are discarded.

  The agent also adds a little prevention to avoid messing the summaries if the
  agent is restarted / killed abruptly. Please, please, please, DO NOT DO IT !


.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN SummarizeLogsAgent
  :end-before: ##END
  :dedent: 2
  :caption: SummarizeLogsAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient

AGENT_NAME = "ResourceStatus/SummarizeLogsAgent"


class SummarizeLogsAgent(AgentModule):
    """SummarizeLogsAgent as extension of AgentModule."""

    def __init__(self, *args, **kwargs):
        """Constructor."""

        AgentModule.__init__(self, *args, **kwargs)

        self.rsClient = None
        self.months = 36

    def initialize(self):
        """Standard initialize.

        :return: S_OK

        """

        self.rsClient = ResourceStatusClient()
        self.months = self.am_getOption("Months", self.months)
        return S_OK()

    def execute(self):
        """execute (main method)

        The execute method runs over the three families of tables (Site, Resource and
        Node) performing identical operations. First, selects all logs for a given
        family (and keeps track of which one is the last row ID). It summarizes the
        logs and finally, deletes the logs from the database.

        At last, this agent removes older entries from history tables

        :return: S_OK
        """

        # loop over the tables
        for element in ("Site", "Resource", "Node"):

            self.log.info("Summarizing %s" % element)

            # get all logs to be summarized
            selectLogElements = self._summarizeLogs(element)
            if not selectLogElements["OK"]:
                self.log.error(selectLogElements["Message"])
                continue

            lastID, logElements = selectLogElements["Value"]

            # logElements is a dictionary of key-value pairs as follows:
            # (name, statusType) : list(logs)
            for key, logs in logElements.items():

                sumResult = self._registerLogs(element, key, logs)
                if not sumResult["OK"]:
                    self.log.error(sumResult["Message"])
                    continue

            if lastID is not None:
                self.log.info("Deleting %sLog till ID %s" % (element, lastID))
                deleteResult = self.rsClient.deleteStatusElement(element, "Log", meta={"older": ["ID", lastID]})
                if not deleteResult["OK"]:
                    self.log.error(deleteResult["Message"])
                    continue

        if self.months:
            self._removeOldHistoryEntries(element, self.months)

        return S_OK()

    def _summarizeLogs(self, element):
        """given an element, selects all logs in table <element>Log.

        :param str element: name of the table family (either Site, Resource or Node)
        :return: S_OK(lastID, listOfLogs) / S_ERROR
        """

        selectResults = self.rsClient.selectStatusElement(element, "Log")

        if not selectResults["OK"]:
            return selectResults

        selectedItems = {}
        latestID = None

        if not selectResults["Value"]:
            return S_OK((latestID, selectedItems))

        selectColumns = selectResults["Columns"]
        selectResults = selectResults["Value"]

        if selectResults:
            latestID = dict(zip(selectColumns, selectResults[-1]))["ID"]

        for selectResult in selectResults:

            elementDict = dict(zip(selectColumns, selectResult))

            key = (elementDict["Name"], elementDict["StatusType"])

            if key not in selectedItems:
                selectedItems[key] = [elementDict]
            else:
                lastStatus = selectedItems[key][-1]["Status"]
                lastToken = selectedItems[key][-1]["TokenOwner"]

                # If there are no changes on the Status or the TokenOwner with respect
                # the previous one, discards the log.
                if lastStatus != elementDict["Status"] or lastToken != elementDict["TokenOwner"]:
                    selectedItems[key].append(elementDict)

        return S_OK((latestID, selectedItems))

    def _registerLogs(self, element, key, logs):
        """Given an element, a key - which is a tuple (<name>, <statusType>)
        and a list of dictionaries, this method inserts them on the <element>History
        table. Before inserting them, checks whether the first one is or is not on
        the <element>History table. If it is, it is not inserted.


        :param str element: name of the table family (either Site, Resource or Node)
        :param tuple key: tuple with the name of the element and the statusType
        :param list logs: list of dictionaries containing the logs
        :return: S_OK(lastID, listOfLogs) / S_ERROR

         :return: S_OK / S_ERROR
        """

        if not logs:
            return S_OK()

        # Undo key
        name, statusType = key

        selectedRes = self.rsClient.selectStatusElement(
            element,
            "History",
            name,
            statusType,
            meta={"columns": ["Status", "TokenOwner"], "limit": 1, "order": ["DateEffective", "desc"]},
        )

        if not selectedRes["OK"]:
            return selectedRes
        selectedRes = selectedRes["Value"]
        if not selectedRes:
            for selectedItemDict in logs:
                res = self.__logToHistoryTable(element, selectedItemDict)
                if not res["OK"]:
                    return res
                return S_OK()

        # We want from the <element>History table the last Status, and TokenOwner
        lastStatus, lastToken = None, None
        if selectedRes:
            try:
                lastStatus = selectedRes[0][0]
                lastToken = selectedRes[0][1]
            except IndexError:
                pass

        # If the first of the selected items has a different status than the latest
        # on the history, we keep it, otherwise we remove it.
        if logs[0]["Status"] == lastStatus and logs[0]["TokenOwner"] == lastToken:
            logs.pop(0)

        if logs:
            self.log.info("%s (%s):" % (name, statusType))
            self.log.debug(logs)

        for selectedItemDict in logs:

            res = self.__logToHistoryTable(element, selectedItemDict)
            if not res["OK"]:
                return res

        return S_OK()

    def __logToHistoryTable(self, element, elementDict):
        """Given an element and a dictionary with all the arguments, this method
        inserts a new entry on the <element>History table

        :param str element: name of the table family (either Site, Resource or Node)
        :param dict elementDict: dictionary returned from the DB to be inserted on the History table

        :return: S_OK / S_ERROR
        """

        name = elementDict.get("Name")
        statusType = elementDict.get("StatusType")
        # vo = elementDict.get('VO')  # FIXME: not sure about it
        status = elementDict.get("Status")
        elementType = elementDict.get("ElementType")
        reason = elementDict.get("Reason")
        dateEffective = elementDict.get("DateEffective")
        lastCheckTime = elementDict.get("LastCheckTime")
        tokenOwner = elementDict.get("TokenOwner")
        tokenExpiration = elementDict.get("TokenExpiration")

        self.log.info("  %s %s %s %s" % (status, dateEffective, tokenOwner, reason))

        return self.rsClient.insertStatusElement(
            element=element,
            tableType="History",
            name=name,
            statusType=statusType,
            status=status,
            elementType=elementType,
            reason=reason,
            dateEffective=dateEffective,
            lastCheckTime=lastCheckTime,
            tokenOwner=tokenOwner,
            tokenExpiration=tokenExpiration,
        )

    def _removeOldHistoryEntries(self, element, months):
        """Delete entries older than period

        :param str element: name of the table family (either Site, Resource or Node)
        :param int months: number of months

        :return: S_OK / S_ERROR
        """
        toRemove = datetime.utcnow().replace(microsecond=0) - timedelta(days=30 * months)
        self.log.info("Removing history entries", "older than %s" % toRemove)

        deleteResult = self.rsClient.deleteStatusElement(
            element, "History", meta={"older": ["DateEffective", toRemove]}
        )
        if not deleteResult["OK"]:
            self.log.error(deleteResult["Message"])



""" MacroCommand

  The MacroCommand class is a macro class for all the macro commands
  for interacting with multiple commands

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger
from DIRAC.ResourceStatusSystem.Command.Command import Command

__RCSID__ = "$Id$"


class MacroCommand(Command):
    """
    As of today, it is not used.
    """

    def __init__(self):

        super(MacroCommand, self).__init__()

        self.commands = None
        self.args = None
        self.clients = None

    ################################################################################

    def setCommands(self, commandsListIn=None):
        """
        Method to be called as first at every MacroCommand instantiation.

        :params:

          :attr:`commandsListIn`: a list of command objects
        """
        if not isinstance(commandsListIn, list):
            commandsListIn = [commandsListIn]

        self.commands = commandsListIn

    ################################################################################

    def setArgs(self, argsListIn=None):
        """
        Set the arguments of the commands.

        :params:

          :attr:`argsListIn`: a list of args tuples, or just a tuple of args
        """
        if not isinstance(argsListIn, list):
            argsListIn = [argsListIn]

        self.args = argsListIn
        commArgs = []

        if len(self.args) == len(self.commands):
            commArgs = [(self.commands[x], self.args[x]) for x in range(len(self.commands))]
        elif len(self.args) == 1:
            commArgs = [(self.commands[x], self.args[0]) for x in range(len(self.commands))]
        else:
            gLogger.error("Tuples or `args` provided are nor 1 nor the same number of the commands")

        for command, arg in commArgs:
            command.setArgs(arg)

    ################################################################################

    def setClient(self, clientListIn=None):
        """
        Set `self.clients`. If not set, a standard client will be instantiated.
        Then, set the clients used by the commands.

        :params:
          :attr:`clientListIn`: a list of client object
        """
        if not isinstance(clientListIn, list):
            clientListIn = [clientListIn]

        self.clients = clientListIn
        commArgs = []

        if len(self.clients) == len(self.commands):
            commArgs = [(self.commands[x], self.clients[x]) for x in range(len(self.commands))]
        elif len(self.args) == 1:
            commArgs = [(self.commands[x], self.clients[0]) for x in range(len(self.commands))]
        else:
            gLogger.error("`clients` provided are nor 1 nor the same number of the commands")

        for command, client in commArgs:
            command.setClient(client)

    ################################################################################

    def doCommand(self):
        """
        Calls command.doCommand for every command in the list of self.commands
        """

        res = []

        for command in self.commands:
            res.append(command.doCommand())
        return res



""" (pytest) unit test for ComponentInstall.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.FrameworkSystem.Client.ComponentInstaller import ComponentInstaller

ci = ComponentInstaller()


def test_getAvailableDatabases():
    res = ci.getAvailableDatabases([])
    assert res["OK"] is True
    assert "JobDB" in res["Value"]
    assert res["Value"]["JobDB"]["System"] == "WorkloadManagement"


def test_getSoftwareComponents():
    res = ci.getSoftwareComponents([])
    assert res["OK"] is True
    assert "Services" in res["Value"]



""" MoveReplica operation handler

    This handler moves replicas from source SEs to target SEs. Replicas are first replicated to target SEs and then removed from the source SEs
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# # imports
import os

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import DMSRequestOperationsBase
from DIRAC.DataManagementSystem.Client.ConsistencyInspector import ConsistencyInspector

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

####


class MoveReplica(DMSRequestOperationsBase):
    """
    .. class:: MoveReplica

    MoveReplica operation handler
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param self: self reference
        :param ~Operation.Operation operation: Operation instance
        :param str csPath: CS path for this handler
        """
        super(MoveReplica, self).__init__(operation, csPath)

        # Init ConsistencyInspector: used to check replicas
        self.ci = ConsistencyInspector()

    def __call__(self):
        """call me maybe"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # own gMonitor stuff for files
            gMonitor.registerActivity(
                "ReplicateAndRegisterAtt",
                "Replicate and register attempted",
                "RequestExecutingAgent",
                "Files/min",
                gMonitor.OP_SUM,
            )
            gMonitor.registerActivity(
                "ReplicateOK", "Replications successful", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "ReplicateFail", "Replications failed", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RegisterOK", "Registrations successful", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RegisterFail", "Registrations failed", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RemoveReplicaAtt", "Replica removals attempted", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RemoveReplicaOK", "Successful replica removals", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RemoveReplicaFail", "Failed replica removals", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )

        # # check replicas first
        res = self.__checkReplicas()
        if not res["OK"]:
            self.log.error("Failed to check replicas", res["Message"])

        sourceSE = self.operation.SourceSE if self.operation.SourceSE else None
        if sourceSE:
            # # check source se for read
            bannedSource = self.checkSEsRSS(sourceSE, "ReadAccess")
            if not bannedSource["OK"]:
                if self.rmsMonitoring:
                    for status in ["Attempted", "Failed"]:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                    self.rmsMonitoringReporter.commit()
                else:
                    gMonitor.addMark("ReplicateAndRegisterAtt", len(self.operation))
                    gMonitor.addMark("ReplicateFail", len(self.operation))
                return bannedSource

            if bannedSource["Value"]:
                self.operation.Error = "SourceSE %s is banned for reading" % sourceSE
                self.log.info(self.operation.Error)
                return S_OK(self.operation.Error)

        # # check targetSEs for write
        bannedTargets = self.checkSEsRSS()
        if not bannedTargets["OK"]:
            if self.rmsMonitoring:
                for status in ["Attempted", "Failed"]:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("ReplicateAndRegisterAtt", len(self.operation))
                gMonitor.addMark("ReplicateFail", len(self.operation))
            return bannedTargets

        if bannedTargets["Value"]:
            self.operation.Error = "%s targets are banned for writing" % ",".join(bannedTargets["Value"])
            return S_OK(self.operation.Error)

        # Can continue now
        self.log.verbose("No targets banned for writing")

        # # check sourceSEs for removal
        # # for removal the targetSEs are the sourceSEs of the replication
        targetSEs = self.operation.sourceSEList
        bannedTargets = self.checkSEsRSS(targetSEs, access="RemoveAccess")
        if not bannedTargets["OK"]:
            if self.rmsMonitoring:
                for status in ["Attempted", "Failed"]:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("RemoveReplicaAtt")
                gMonitor.addMark("RemoveReplicaFail")
            return bannedTargets

        if bannedTargets["Value"]:
            return S_OK("%s targets are banned for removal" % ",".join(bannedTargets["Value"]))

        # Can continue now
        self.log.verbose("No targets banned for removal")

        # Do the transfer
        # # get waiting files. If none just return
        waitingFiles = self.getWaitingFilesList()
        if not waitingFiles:
            return S_OK()

        # # loop over files
        self.log.info("Transferring files using Data manager...")
        for opFile in waitingFiles:
            res = self.dmTransfer(opFile)
            if not res["OK"]:
                continue
            else:
                # Do the replica removal
                self.log.info("Removing files using Data manager...")
                toRemoveDict = dict([(opFile.LFN, opFile) for opFile in waitingFiles])
                self.log.info("todo: %s replicas to delete from %s sites" % (len(toRemoveDict), len(targetSEs)))
                self.dmRemoval(toRemoveDict, targetSEs)

        return S_OK()

    def __checkReplicas(self):
        """check done replicas and update file states"""
        waitingFiles = dict(
            [(opFile.LFN, opFile) for opFile in self.operation if opFile.Status in ("Waiting", "Scheduled")]
        )
        targetSESet = set(self.operation.targetSEList)

        # Check replicas
        res = self.ci._getCatalogReplicas(list(waitingFiles))

        if not res["OK"]:
            self.log.error("Failed to get catalog replicas", res["Message"])
            return S_ERROR()

        allReplicas = res["Value"][0]

        replicas = self.ci.compareChecksum(list(waitingFiles))

        if not replicas["OK"]:
            self.log.error("Failed to check replicas", replicas["Message"])
            return S_ERROR()

        replicas = replicas["Value"]
        noReplicas = replicas["NoReplicas"]

        if noReplicas:
            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", len(noReplicas)))
                self.rmsMonitoringReporter.commit()
            for lfn in noReplicas.keys():
                self.log.error("File %s doesn't exist" % lfn)
                if not self.rmsMonitoring:
                    gMonitor.addMark("ReplicateFail", len(targetSESet))
                waitingFiles[lfn].Status = "Failed"

        for lfn, reps in allReplicas.items():
            if targetSESet.issubset(set(reps)):
                self.log.info("file %s has been replicated to all targets" % lfn)
                waitingFiles[lfn].Status = "Done"

        return S_OK()

    def dmRemoval(self, toRemoveDict, targetSEs):

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(toRemoveDict)))
            self.rmsMonitoringReporter.commit()
        else:
            gMonitor.addMark("RemoveReplicaAtt", len(toRemoveDict) * len(targetSEs))
        # # keep status for each targetSE
        removalStatus = dict.fromkeys(toRemoveDict.keys(), None)
        for lfn in removalStatus:
            removalStatus[lfn] = dict.fromkeys(targetSEs, None)

        # # loop over targetSEs
        for targetSE in targetSEs:
            self.log.info("removing replicas at %s" % targetSE)

            # # 1st step - bulk removal
            bulkRemoval = self.bulkRemoval(toRemoveDict, targetSE)
            if not bulkRemoval["OK"]:
                self.log.error("Bulk replica removal failed", bulkRemoval["Message"])
                return bulkRemoval
            bulkRemoval = bulkRemoval["Value"]

            # # update removal status for successful files
            removalOK = [opFile for opFile in bulkRemoval.values() if not opFile.Error]

            for opFile in removalOK:
                removalStatus[opFile.LFN][targetSE] = ""

            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", len(removalOK)))
            else:
                gMonitor.addMark("RemoveReplicaOK", len(removalOK))

            # # 2nd step - process the rest again
            toRetry = dict([(lfn, opFile) for lfn, opFile in bulkRemoval.items() if opFile.Error])
            for lfn, opFile in toRetry.items():
                self.singleRemoval(opFile, targetSE)
                if not opFile.Error:
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
                    else:
                        gMonitor.addMark("RemoveReplicaOK", 1)
                    removalStatus[lfn][targetSE] = ""
                else:
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                    else:
                        gMonitor.addMark("RemoveReplicaFail", 1)
                    removalStatus[lfn][targetSE] = opFile.Error

        # # update file status for waiting files
        failed = 0
        for opFile in self.operation:
            if opFile.Status == "Waiting":
                errors = list(set([error for error in removalStatus[lfn].values() if error]))
                if errors:
                    opFile.Error = ",".join(errors)
                    # This seems to be the only offending error
                    if "Write access not permitted for this credential" in opFile.Error:
                        failed += 1
                        continue
                opFile.Status = "Done"

        if failed:
            self.operation.Error = "failed to remove %s replicas" % failed

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        return S_OK(removalStatus)

    def dmTransfer(self, opFile):
        """replicate and register using dataManager"""
        # # get waiting files. If none just return
        # # source SE
        sourceSE = self.operation.SourceSE if self.operation.SourceSE else None

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", 1))
            self.rmsMonitoringReporter.commit()
        else:
            gMonitor.addMark("ReplicateAndRegisterAtt", 1)

        opFile.Error = ""
        lfn = opFile.LFN

        # Check replicas
        res = self.ci._getCatalogReplicas([lfn])

        if not res["OK"]:
            self.log.error("Failed to get catalog replicas", res["Message"])
            return S_ERROR()

        allReplicas = res["Value"][0]
        replicas = self.ci.compareChecksum([lfn])

        if not replicas["OK"]:
            self.log.error("Failed to check replicas", replicas["Message"])
            return S_ERROR()

        replicas = replicas["Value"]

        validReplicas = []
        noReplicas = replicas["NoReplicas"]
        missingAllReplicas = replicas["MissingAllReplicas"]
        missingReplica = replicas["MissingReplica"]
        someReplicasCorrupted = replicas["SomeReplicasCorrupted"]
        allReplicasCorrupted = replicas["AllReplicasCorrupted"]

        if noReplicas:
            self.log.error("Unable to replicate", "File %s doesn't exist" % (lfn))
            opFile.Error = "No replicas found"
            opFile.Status = "Failed"
        elif missingAllReplicas:
            self.log.error("Unable to replicate", "%s, all replicas are missing" % (lfn))
            opFile.Error = "Missing all replicas"
            opFile.Status = "Failed"
        elif allReplicasCorrupted:
            self.log.error("Unable to replicate", "%s, all replicas are corrupted" % (lfn))
            opFile.Error = "All replicas corrupted"
            opFile.Status = "Failed"
        elif someReplicasCorrupted:
            self.log.error("Unable to replicate", "%s, replicas corrupted at %s" % (lfn, someReplicasCorrupted[lfn]))
            opFile.Error = "At least one replica corrupted"
            opFile.Status = "Failed"
        elif missingReplica:
            self.log.error("Unable to replicate", "%s, missing replicas at %s" % (lfn, missingReplica[lfn]))
            opFile.Error = "At least one missing replica"
            opFile.Status = "Failed"

        if opFile.Error:
            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("ReplicateFail")
            return S_ERROR()

        # Check if replica is at the specified source
        for repSEName in allReplicas[lfn]:
            validReplicas.append(repSEName)

        # # get the first one in the list
        if sourceSE not in validReplicas:
            if sourceSE:
                self.log.warn("%s is not at specified sourceSE %s, changed to %s" % (lfn, sourceSE, validReplicas[0]))
            sourceSE = validReplicas[0]

        # # loop over targetSE
        catalogs = self.operation.Catalog
        if catalogs:
            catalogs = [cat.strip() for cat in catalogs.split(",")]

        for targetSE in self.operation.targetSEList:
            # # call DataManager
            if targetSE in validReplicas:
                self.log.warn("Request to replicate %s to an existing location: %s" % (lfn, targetSE))
                continue

            res = self.dm.replicateAndRegister(lfn, targetSE, sourceSE=sourceSE, catalog=catalogs)

            if res["OK"]:
                if lfn in res["Value"]["Successful"]:
                    if "replicate" in res["Value"]["Successful"][lfn]:
                        repTime = res["Value"]["Successful"][lfn]["replicate"]
                        prString = "file %s replicated at %s in %s s." % (lfn, targetSE, repTime)

                        if not self.rmsMonitoring:
                            gMonitor.addMark("ReplicateOK", 1)

                        if "register" in res["Value"]["Successful"][lfn]:

                            if not self.rmsMonitoring:
                                gMonitor.addMark("RegisterOK", 1)

                            regTime = res["Value"]["Successful"][lfn]["register"]
                            prString += " and registered in %s s." % regTime
                            self.log.info(prString)
                        else:

                            if not self.rmsMonitoring:
                                gMonitor.addMark("RegisterFail", 1)

                            prString += " but failed to register"
                            self.log.warn(prString)

                            opFile.Error = "Failed to register"
                            # # add register replica operation
                            registerOperation = self.getRegisterOperation(opFile, targetSE, type="RegisterReplica")
                            self.request.insertAfter(registerOperation, self.operation)
                    else:
                        self.log.error("Failed to replicate", "%s to %s" % (lfn, targetSE))

                        if not self.rmsMonitoring:
                            gMonitor.addMark("ReplicateFail", 1)

                        opFile.Error = "Failed to replicate"
                else:

                    if not self.rmsMonitoring:
                        gMonitor.addMark("ReplicateFail", 1)

                    reason = res["Value"]["Failed"][lfn]
                    self.log.error("Failed to replicate and register", "File %s at %s: %s" % (lfn, targetSE, reason))
                    opFile.Error = reason
            else:

                if not self.rmsMonitoring:
                    gMonitor.addMark("ReplicateFail", 1)

                opFile.Error = "DataManager error: %s" % res["Message"]
                self.log.error("DataManager error", res["Message"])

        if not opFile.Error:
            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
            if len(self.operation.targetSEList) > 1:
                self.log.info("file %s has been replicated to all targetSEs" % lfn)
        else:
            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                self.rmsMonitoringReporter.commit()
            return S_ERROR("dmTransfer failed")

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        return S_OK()

    def bulkRemoval(self, toRemoveDict, targetSE):
        """remove replicas :toRemoveDict: at :targetSE:

        :param dict toRemoveDict: { lfn: opFile, ... }
        :param str targetSE: target SE name
        :return: toRemoveDict with updated errors
        """
        removeReplicas = self.dm.removeReplica(targetSE, list(toRemoveDict))

        if not removeReplicas["OK"]:
            for opFile in toRemoveDict.values():
                opFile.Error = removeReplicas["Message"]
            return S_ERROR(removeReplicas["Message"])
        removeReplicas = removeReplicas["Value"]
        # # filter out failed
        for lfn, opFile in toRemoveDict.items():
            if lfn in removeReplicas["Failed"]:
                opFile.Error = str(removeReplicas["Failed"][lfn])
        return S_OK(toRemoveDict)

    def singleRemoval(self, opFile, targetSE):
        """remove opFile replica from targetSE

        :param ~DIRAC.RequestManagementSystem.Client.File.File opFile: File instance
        :param str targetSE: target SE name
        """
        proxyFile = None
        if "Write access not permitted for this credential" in opFile.Error:
            # # not a DataManger? set status to failed and return
            if "DataManager" in self.shifter:
                # #  you're a data manager - save current proxy and get a new one for LFN and retry
                saveProxy = os.environ["X509_USER_PROXY"]
                try:
                    fileProxy = self.getProxyForLFN(opFile.LFN)
                    if not fileProxy["OK"]:
                        opFile.Error = fileProxy["Message"]
                    else:
                        proxyFile = fileProxy["Value"]
                        removeReplica = self.dm.removeReplica(targetSE, opFile.LFN)
                        if not removeReplica["OK"]:
                            opFile.Error = removeReplica["Message"]
                        else:
                            removeReplica = removeReplica["Value"]
                            if opFile.LFN in removeReplica["Failed"]:
                                opFile.Error = removeReplica["Failed"][opFile.LFN]
                            else:
                                # # reset error - replica has been removed this time
                                opFile.Error = ""
                finally:
                    if proxyFile:
                        os.unlink(proxyFile)
                    # # put back request owner proxy to env
                    os.environ["X509_USER_PROXY"] = saveProxy
        return S_OK(opFile)
